//"use strict";

// Names used in chessboard.css .
var pieceNames = ["pawn", "knight", "bishop", "rook", "queen", "king",
                  "nightrider", "berolina", "fers", "wazir", "erlking",
                  "princess", "empress", "amazon", "waran", "elephant",
                  "dragon", "ship"];

/* piece test

englishN
forsyth 8/pnbrqkhu/fwmacvje/ds6/8/PNBRQKHU/FWMACVJE/DS6
play help 0.5

*/

// ---------------------------------------------------------------------------
// Try-catch wrapper.

function Viewer_ComputeLegalMovesSafe() {
	var initialStack = Module.stackSave();
	try {
		return Viewer_ComputeLegalMoves();
	}
	catch (e) {
		catchFn(e);
		Module.stackRestore(initialStack);
		return "";
	}
}

// ---------------------------------------------------------------------------
// Game editor moves.

function applyEditableCheckbox() {
	var checked = $("input#editable-game").is(":checked");

	if (checked) {
		$("p#game-moves").hide();
		$("form#editable-moves").show();
		$("form#editable-moves textarea").keyup();	// Fix textarea height.

	}
	else {
		$("form#editable-moves").hide();
		$("p#game-moves").show();
	}
}

// Call this when the game changes.
function updateGameMoves() {
	// Always update both move displays, even though only one of them is
	// shown at the same time.
	var langObj = languageObject();
	$("p#game-moves").html(
		solutionToHtml(g_editedGame, langObj, 2) +
		"<em style='font-style: normal; color: red'>" + g_editedGame.reject +
		"</em>");

	// Do not modify textarea if we have nothing to put in it, in case
	// what's already in it is more interesting.
	if (g_editedGame.moves.length > 0 || g_editedGame.reject != "") {
		var $textarea = $("form#editable-moves textarea");
		$textarea.val(
			solutionToHtml(g_editedGame, langObj, 0) + g_editedGame.reject);
		// Only call keyup if the textarea is visible.
		var checked = $("input#editable-game").is(":checked");
		if (checked)
			$textarea.keyup();
	}
}

// ---------------------------------------------------------------------------
// Game editor position.

function playerMustPass(legalMoves, len) {
	return legalMoves.length == len && legalMoves.substr(0, 5) == "-----";
}

// Automatically play obligatory pass moves.
// Store the list of legal moves.
function autoPlayPassMoves() {
	var len = Viewer_GetMoveLength();
	while (true) {
		g_legalMoves = Viewer_ComputeLegalMovesSafe();
		if (playerMustPass(g_legalMoves, len)) {
			SP_PlayMove(VIEWER, 0);
			if (g_editedDepth == g_editedGame.moves.length) {
				// Adding a new pass at the end of the game.
				g_editedGame.moves[g_editedDepth] = getLastMove(VIEWER);
				// No move number for the last pass.
				g_editedGame.moves[g_editedDepth].str = "";
			}
			g_editedDepth++;
		}
		else
			break;
	}
}

function pieceSize() {
//	return wantMini() ? 34 : 45;

	// To get a sharp board, we use td's instead of a board image (which would
	// become blurry if scaled by a factional number). The price to pay is
	// that the css-size of td's is not exactly what we asked for, and might
	// not even be a square!
	// Experimentally measure the piece size.
	var offset = $("td#game-board").offset();
	var offset2 = $("td#game-board2").offset();
	return {x: (offset2.left - offset.left) / 8,
	        y: (offset2.top - offset.top) / 8};
}

function updateBoard() {
	// Extract all 64 squares.
	var squares = [];
	for (var i = 0; i < 64; i++)
		squares.push(Viewer_GetSquare(i));

	// Create a list of possible source squares.
	var srcList = [];
	var len = Viewer_GetMoveLength();
	for (var i = 0; i < g_legalMoves.length; i += len) {
		var srcSquare = g_legalMoves.substr(i, 2);
		if (srcList.indexOf(srcSquare) == -1)
			srcList.push(srcSquare);
	}

	var $board = $("td#game-board");
	var size = pieceSize();

	$board.html("");
	g_moveState = 0;	// Highlight is gone.

	var whiteCount = 0;
	var blackCount = 0;
	for (var y = 0; y < 8; y++) {
		for (var x = 0; x < 8; x++) {
			var v = squares[x + y*8];
			if (v == EMPTY)
				continue;
			var piece = v & (WHITE - 1);
			if (piece >= pieceNames.length)
				openDialog("Bug", "Unknown piece type.");
			var pieceStr = pieceNames[piece];

			var colorStr;
			if ((v & WHITE) != 0) {
				colorStr = "white";
				whiteCount++;
			}
			else if ((v & BLACK) != 0) {
				colorStr = "black";
				blackCount++;
			}
			else
				openDialog("Bug", "Unknown piece color.");

			var royal_flag = (v & ROYAL) != 0;

			var square = String.fromCharCode(97 + x, 56 - y);
			var draggable = (srcList.indexOf(square) != -1) ? "draggable " : "";

			$board.append("<div class='" + square + " piece " + draggable +
			              colorStr + " " + pieceStr +
			              "' " + "style='position:absolute; " +
			              "left:" + (x*size.x).toFixed(3) + "px; " +
			              "top:" + (y*size.y).toFixed(3) + "px;'></div>");
			if (royal_flag) {
				$board.append("<div class='" + square + " piece " + draggable +
				              "royal" +
				              "' " + "style='position:absolute; " +
				              "left:" + (x*size.x).toFixed(3) + "px; " +
				              "top:" + (y*size.y).toFixed(3) + "px;'></div>");
			}
		}
	}

	$("span#game-piece-counts").text(
		"(" + whiteCount.toString() + "+" + blackCount.toString() + ")");
}

function buildForsyth(squares, langObj) {
	// Figurine forsyth is not useful so use englishS letters instead.
	var letters = (langObj.lang == "figurine") ?
	              pieceLetters.englishS : langObj.letters;

	var ret = "forsyth ";
	var count = 0;
	for (var i = 0; i < 64; i++) {
		var v = squares[i];
		if (v == EMPTY)
			count++;
		else if ((v & ROYAL) != 0)	// No way to set royal in forsyth.
			return "";
		else {
			if (count > 0) {
				ret += count.toString();
				count = 0;
			}
			var piece = letters[v & TYPE];
			if (piece.length != 1)
				ret += ".";	// A dot announces a 2-letter piece.
			if ((v & WHITE) != 0)
				ret += piece;
			else if ((v & BLACK) != 0)
				ret += piece.toLowerCase();
			else
				return "";
		}

		if (i % 8 == 7) {
			if (count > 0) {
				ret += count.toString();
				count = 0;
			}
			if (i < 63)
				ret += "/";
		}
	}
	return ret;
}

function buildPieces(squares, langObj) {
	// Figurine pieces are not useful so use englishS letters instead.
	var letters = (langObj.lang == "figurine") ?
	              pieceLetters.englishS : langObj.letters;

	var ret = langObj.keywords[0];
	// List pieces in the order KQRBSP, then fairy pieces.
	var order = [5, 4, 3, 2, 1, 0];
	for (var i = 6; i < letters.length - 1; i++)	// Exclude 'X'.
		order.push(i);

	for (var i = 0; i < 4; i++) {
		var color = [WHITE + ROYAL, WHITE, BLACK + ROYAL, BLACK][i];

		var keywords = langObj.keywords[(i >> 1) + 1];	// White or black.
		if ((i & 1) == 0)
			keywords += " " + langObj.keywords[3];	// Royal.

		var declareColor = true;
		for (var j = 0; j < order.length; j++) {
			var piece = order[j];
			var declarePiece = true;
			for (var y = 7; y >= 0; y--) {
				for (var x = 0; x < 8; x++) {
					var v = squares[x + y*8];
					if (v == color + piece) {
						if (declareColor) {
							ret += " " + keywords;
							declareColor = false;
						}
						if (declarePiece) {
							ret += " " + letters[piece];
							declarePiece = false;
						}
						var file = String.fromCharCode(97 + x);
						var rank = String.fromCharCode(56 - y);
						if (langObj.lang == "french2") {
							if (file == "c")
								file = "\u00E7";
							else if (file == "e")
								file = "\u00E9";
						}
						ret += file + rank;
					}
				}
			}
		}
	}

	return ret;
}

// ---------------------------------------------------------------------------

// Call this (and updateBoard) when the selected move within the game changes.
function updateSelectedMove() {
	// Find previous depth, or -1.
	var prev_depth = g_editedDepth - 1;
	while (prev_depth >= 0) {
		var internal = g_editedGame.moves[prev_depth].internal;
		if (internal.substr(0, 5) != "-----")
			break;
		prev_depth--;
	}

	// Erase previously highlighted move.
	$("p#game-moves > span").css("color", "")
	                        .css("background-color", "");

	// Highlight move.
	$("p#game-moves > span." +
	  (prev_depth == -1 ? "initial" : prev_depth))
		.css("color", "white")
		.css("background-color", "black");

	// Also update forsyth and pieces so update is instant and doesn't wait
	// for animation.
	// Extract all 64 squares.
	var squares = [];
	for (var i = 0; i < 64; i++)
		squares.push(Viewer_GetSquare(i));

	var langObj = languageObject();
	$("p#game-forsyth").html(buildForsyth(squares, langObj));
	$("p#game-pieces").html(buildPieces(squares, langObj));
}

// Fix the game editor after the game has changed.
// Precondition: the viewer is set to the end position.
// Walk backwards until g_editedDepth, then update board and moves.
function finalizeGameEditor() {
	updateGameMoves();

	if (g_editedDepth > g_editedGame.moves.length)
		g_editedDepth = g_editedGame.moves.length;

	// Reverse until we're back to our original depth.
	for (var i = g_editedDepth; i < g_editedGame.moves.length; i++)
		SP_ReverseMove(VIEWER);

	autoPlayPassMoves();
	updateBoard();
	updateSelectedMove();
}

function reduceMove(move) {
	// Grab the first 4 characters, and the first capital letter
	// among characters 4-7, if any.
	var ret = move.substr(0, 4);
	for (var i = 4; i < 8; i++) {
		var c = move.charAt(i);
		if (c >= "A" && c <= "Z") {
			ret += c;
			break;
		}
	}
	return ret;
}

function findMoveMatch(legalMoves, len, move) {
	var move2 = reduceMove(move);
	var k2 = -1;
	for (var k = 0; k * len < legalMoves.length; k++) {
		var candidate = legalMoves.substr(k*len, len);
		if (candidate == move)
			return k;

		// As a second choice, try a reduced match if the match is unique.
		if (reduceMove(candidate) == move2) {
			if (k2 == -1)
				k2 = k;
			else
				k2 = -2;
		}
	}
	return k2;
}

// Called when the start button is clicked. We attempt to use the edited
// game with the new conditions. Also called by copyGameToEditor, where the
// attempt to use the edited game should always work fully.
// Precondition: the viewer is set to the initial position, but g_editedGame
// might contain a game and g_editedDepth might be nonzero.
function initGameEditor() {
	// We attempt to use as much of the previous edited game as possible.
	// This step is similar to extractEditedGame, but much simpler because
	// we have access to the internal move representation.

	if (g_editedGame.row960 != "") {
		// Try to set Row960.
		if (!Viewer_SetRow960(g_editedGame.row960)) {
			// If it failed, then make everything red.
			var reject = g_editedGame.str960;
			for (var j = 0; j < g_editedGame.moves.length; j++)
				reject += g_editedGame.moves[j].str;
			reject += g_editedGame.reject;

			// Erase the game.
			g_editedGame.row960 = "";
			g_editedGame.moves.length = 0;
			g_editedGame.reject = reject;

			finalizeGameEditor();
			return;
		}
	}

	var len = Viewer_GetMoveLength();
	// Play until the end.
	for (var i = 0; i < g_editedGame.moves.length; i++) {
		var legalMoves = Viewer_ComputeLegalMovesSafe();
		var k = findMoveMatch(legalMoves, len,
		                      g_editedGame.moves[i].internal);
		if (k < 0) {
			// Can't find the move. Put the rest in red.
			var reject = "";
			for (var j = i; j < g_editedGame.moves.length; j++)
				reject += g_editedGame.moves[j].str;
			reject += g_editedGame.reject;

			// Truncate game here.
			g_editedGame.moves.length = i;
			g_editedGame.reject = reject;
			break;
		}
		SP_PlayMove(VIEWER, k);
		g_editedGame.moves[i] = getLastMove(VIEWER);
	}

	finalizeGameEditor();
}

// Called when a solution is clicked.
function copyGameToEditor(game, span) {
	// Reverse the moves played so far.
	for (var i = 0; i < g_editedDepth; i++)
		SP_ReverseMove(VIEWER);

	// Deep clone the game object.
	g_editedGame = JSON.parse(JSON.stringify(game));

	// Remove old highlight.
	g_editedSpan.css("background-color", "");

	// Add new highlight.
	g_editedSpan = $(span);
	g_editedSpan.css("background-color", "#ffc");

	g_editedGame.reject = "";

	// Clear the string version of the game to be safe.
	// This shouldn't matter since initGameEditor is supposed to accept
	// our game.
	g_editedGame.str960 = "";
	for (var i = 0; i < g_editedGame.moves.length; i++)
		g_editedGame.moves[i].str = "";

	initGameEditor();
}

function animatePiece($piece, u0, v0, u1, v1) {
	var size = pieceSize();
	$piece.css({
			left: (u0 * size.x).toFixed(3) + "px",
			top: (v0 * size.y).toFixed(3) + "px",
			"z-index": 1
		});

	if (g_verticalCylinder && Math.abs(u1 - u0) > 4) {
		// Fancy animation if going the other direction in vertical cylinder.
		var ua, ub, frac;
		if (u1 > u0) {
			ua = -0.5
			ub = 7.5
			frac = (u0 - ua) / (u0 - u1 + 8);
		}
		else {
			ua = 7.5
			ub = -0.5
			frac = (ua - u0) / (u1 - u0 + 8);
		}
		var va = v0 + frac * (v1 - v0);

		$piece.animate({
				left: (ua * size.x).toFixed(3) + "px",
				top: (va * size.y).toFixed(3) + "px"
			},
			200 * frac, "swing", function() {

		$piece.css({
				left: (ub * size.x).toFixed(3) + "px",
				top: (va * size.y).toFixed(3) + "px"
			});

		$piece.animate({
				left: (u1 * size.x).toFixed(3) + "px",
				top: (v1 * size.y).toFixed(3) + "px"
			},
			200 * (1 - frac), "swing", updateBoard)

			});
	}
	else {
		$piece.animate({
				left: (u1 * size.x).toFixed(3) + "px",
				top: (v1 * size.y).toFixed(3) + "px"
			},
			200, "swing", updateBoard);
	}
}

// Called after the user plays an arbitrary legal move.
function playMove(k) {
	SP_PlayMove(VIEWER, k);
	g_editedGame.moves[g_editedDepth] = getLastMove(VIEWER);
	g_editedDepth++;
	g_editedGame.moves.length = g_editedDepth;
	g_editedGame.reject = "";
	g_editedSpan.css("background-color", "");

	updateGameMoves();

	autoPlayPassMoves();
	updateBoard();
	updateSelectedMove();
}

// If we encounter the characters "{", "[" or "(" without a closing match, then
// the closing character we're waiting for is called "closingChar". This
// function returns the new value of closingChar after going through a token.
function updateClosingChar(closingChar, token) {
	for (var i = 0; i < token.length; i++) {
		var c = token.charAt(i);
		if (closingChar != "") {
			if (c == closingChar)
				closingChar = "";
		}
		else if (c == "{")
			closingChar = "}";
		else if (c == "[")
			closingChar = "]";
		else if (c == "(")
			closingChar = ")";
	}
	return closingChar;
}

// Get rid of optional move markings.
function canonizeMove(move) {
	// Get rid of a brace-type comment, if any.
	var k = move.indexOf("{");
	if (k >= 0) {
		if (k > 0 && move.charAt(k-1) == " ")
			k--;
		move = move.substr(0, k);
	}

	// Remove check, mate, stalemate and auto.
	// Normally they appear at the end, but look for it inside of the move too.
	var look = true;
	for (var i = move.length - 1; i >= 0; i--) {
		var c = move.charAt(i);
		if (look && (c == "+" || c == "#" || c == "=" || c == "!" ||
		             c == "\u2021")) {	// Double dagger.
			move = move.substr(0, i) + move.substr(i+1);
		}
		else
			look = (c == " " || c == "{" || c == "[" || c == "(");
	}

	// Get rid of e.p. Give priority to removing a leading space and a trailing
	// dot.
	var list = [" e.p.", " e.p", " ep.", " ep", "e.p.", "e.p", "ep.", "ep"];
	for (var i = 0; i < list.length; i++) {
		var k = move.indexOf(list[i]);
		if (k >= 0) {
			// Remove e.p. substring.
			move = move.substr(0, k) + move.substr(k + list[i].length);
			break;
		}
	}

	// Standardize zero into O for castling.
	if (move.substr(0, 7) == "0-0-0-0")
		move = "O-O-O-O" + move.substr(7);
	else if (move.substr(0, 5) == "0-0-0")
		move = "O-O-O" + move.substr(5);
	else if (move.substr(0, 3) == "0-0")
		move = "O-O" + move.substr(3);

	// Standardize capture symbols to "x".
	move = move.replace("\u00D7", "x");		// Multiplication sign.
	move = move.replace(":", "x");

	// Standardize accented letters (global replace because of "Rcc1").
	move = move.replace(/\u00E7/g, "c");		// c-cedilla.
	move = move.replace(/\u00E9/g, "e");		// e-acute.

	return move;
}

// Get rid of move effect like "[e1=rS]" which can be annotated differently
// like "(wK=rS)". This may lose promotion information, so it's not part of
// canonizeMove, but instead used as a fallback.
function simplifyMove(move) {
	return move.replace(/[ \[\(].*/, "");
}

// After the call, the viewer is set to the end position, which is out-of-sync
// with g_editedDepth. You must then call finalizeGameEditor to fix this.
function extractEditedGame() {
	var langObj = languageObject();
	var rest = $("form#editable-moves textarea").val();

	// Reverse the moves played so far.
	for (var i = 0; i < g_editedDepth; i++)
		SP_ReverseMove(VIEWER);

	g_editedGame.row960 = "";
	g_editedGame.moves = [];

	var re = /^\s*\((\S{8})\)(.*)$/;
	var results = re.exec(rest);
	if (results) {
		// Translate row960 to englishN.
		for (var i = 0; i < 8; i++) {
			var piece = langObj.letters.indexOf(results[1][i]);
			if (piece >= 0)
				g_editedGame.row960 += pieceLetters.englishN[piece];
		}
		if (!Viewer_SetRow960(g_editedGame.row960)) {
			// Failure.
			g_editedGame.row960 = "";
			g_editedGame.reject = rest;
			return;
		}
		rest = results[2];
	}

	while (true) {
		// If we fail to find the move, then reject will be this.
		g_editedGame.reject = rest;

		// Discard white space, if any. [\s\S] is like "." but works multiline.
		re = /^\s*([\s\S]*)$/;
		results = re.exec(rest);
		rest = results[1];	// This one should always match.

		// Discard move number, if any.
		re = /^-?\d+\.+\s*([\s\S]*)$/;
		results = re.exec(rest);
		if (results)
			rest = results[1];

		// Grab a token.
		re = /^(\S+)([\s\S]*)$/;
		results = re.exec(rest);
		if (!results)
			break;

		var moveStr = results[1];
		rest = results[2];

		var closingChar = updateClosingChar("", moveStr);

		// Peek at the next token.
		while (true) {
			re = /^\s*(\S+)([\s\S]*)$/;
			var results2 = re.exec(rest);
			if (!results2)
				break;
			var peek = results2[1];
			if (closingChar != "") {
				// Include next token because the previous token had an
				// open pair.
				moveStr += " " + peek;
				rest = results2[2];
				closingChar = updateClosingChar(closingChar, peek);
				// Keep going until we hit the closing character.
			}
			else if (peek.substr(0, 2) == "ep" || peek.substr(0, 3) == "e.p" ||
			         peek.charAt(0) == "{" || peek.charAt(0) == "[" ||
			         peek.charAt(0) == "(" ||
			         peek.charAt(0) == "+" || peek.charAt(0) == "#") {
				// Incorporate enpassant, bracket tokens, and checks separated
				// by a space (like by py2web).
				moveStr += " " + peek;
				rest = results2[2];
				closingChar = updateClosingChar(closingChar, peek);
			}
			else
				break;
		}

		moveStr = canonizeMove(moveStr);

		// Find the list of legal moves, skipping passes.
		var legalMoves;
		var len = Viewer_GetMoveLength();
		while (true) {
			legalMoves = Viewer_ComputeLegalMovesSafe();
			if (playerMustPass(legalMoves, len)) {
				SP_PlayMove(VIEWER, 0);
				g_editedGame.moves.push(getLastMove(VIEWER));
			}
			else
				break;
		}

		var simplifiedMatches = [];

		// Try to find the move.
		var k;
		for (k = 0; k * len < legalMoves.length; k++) {
			SP_PlayMove(VIEWER, k);
			// Don't compute check for speed.
			var moveObj = getLastMoveFast(VIEWER);
			var engineStr = canonizeMove(moveToText(moveObj, langObj));
			if (engineStr == moveStr) {
				// We have a match.
				moveObj.check = SP_GetLastMoveNotation(VIEWER, 5);
				g_editedGame.moves.push(moveObj);
				break;
			}
			else if (simplifyMove(engineStr) == simplifyMove(moveStr))
				simplifiedMatches.push(k);
			SP_ReverseMove(VIEWER);
		}
		if (k * len == legalMoves.length) {
			// No canonical match. Use a simplified match if unique.
			if (simplifiedMatches.length == 1) {
				SP_PlayMove(VIEWER, simplifiedMatches[0]);
				var moveObj = getLastMove(VIEWER);
				g_editedGame.moves.push(moveObj);
			}
			else {
				// Move wasn't found or simplified match is ambiguous.
				// g_editedGame.reject is already set.
				return;
			}
		}
	}
	g_editedGame.reject = "";	// Everything was extracted with no error.
}

function resetOriginPiece(size) {
	var u0 = g_originSquare.charCodeAt(0) - 97;
	var v0 = 56 - g_originSquare.charCodeAt(1);

	g_originPiece.css({
			left: (u0 * size.x).toFixed(3) + "px",
			top: (v0 * size.y).toFixed(3) + "px",
			"z-index": 0
		});
	if (g_originPiece.hasClass("dragging")) {
		g_originPiece.removeClass("dragging");
		g_originPiece.addClass("draggable");
	}
}

function cancelMoveChoice() {
	var $select = $("select#move-choice");
	if ($select.length > 0) {
		$select.remove();
		var size = pieceSize();
		resetOriginPiece(size);
	}
}

// value: 1: green, 2: orange, 4: red, other: mix
function addSquareToList(list, square, value) {
	for (var i = 0; i < list.length; i++) {
		if (list[i][0] == square) {
			list[i][1] |= value;
			return;
		}
	}
	list.push([square, value]);
}

$(document).one("pagecreate", "#solver", function() {
	var $page = $(this);	// "this" is the page for pagecreate.

	$page.find("select#output-language").on("change", function(event, ui) {
		// Rewrite game and current position with new language.
		// Make sure that the engine is initialized.
		if (g_editorReady && !ABORT) {
			updateGameMoves();
			updateSelectedMove();	// To fix forsyth and pieces letters.
		}
	});

	$page.find("button#game-start").on("vclick", function(event, ui) {
		cancelMoveChoice();

		if (g_editedDepth > 0 && !ABORT) {
			for (var i = 0; i < g_editedDepth; i++)
				SP_ReverseMove(VIEWER);
			g_editedDepth = 0;
			autoPlayPassMoves();
			updateBoard();
			updateSelectedMove();
		}
	});

	$page.find("button#game-back").on("vclick", function(event, ui) {
		cancelMoveChoice();

		if (g_editedDepth > 0 && !ABORT) {
			SP_ReverseMove(VIEWER);
			g_editedDepth--;

			// Find previous move, skipping pass moves if needed.			
			var internal;
			while (true) {
				internal = g_editedGame.moves[g_editedDepth].internal;
				if (g_editedDepth > 0 && internal.substr(0, 5) == "-----") {
					SP_ReverseMove(VIEWER);
					g_editedDepth--;
				}
				else
					break;
			}

			g_legalMoves = Viewer_ComputeLegalMovesSafe();
			updateBoard();
			updateSelectedMove();

			var startSquare = internal.substr(0, 2);
			var $piece = $("td#game-board > div." + startSquare);
			var u0 = internal.charCodeAt(0) - 97;
			var v0 = 56 - internal.charCodeAt(1);
			var u1 = internal.charCodeAt(2) - 97;
			var v1 = 56 - internal.charCodeAt(3);

			// Animate the piece backwards.
			animatePiece($piece, u1, v1, u0, v0);
		}
	});

	$page.find("button#game-forward").on("vclick", function(event, ui) {
		cancelMoveChoice();

		if (g_editedDepth < g_editedGame.moves.length && !ABORT) {
			var internal = g_editedGame.moves[g_editedDepth].internal;
			SP_PlayMove(VIEWER, -1);
			g_editedDepth++;

			autoPlayPassMoves();
			updateSelectedMove();

			var startSquare = internal.substr(0, 2);
			var $piece = $("td#game-board > div." + startSquare);
			var u0 = internal.charCodeAt(0) - 97;
			var v0 = 56 - internal.charCodeAt(1);
			var u1 = internal.charCodeAt(2) - 97;
			var v1 = 56 - internal.charCodeAt(3);

			animatePiece($piece, u0, v0, u1, v1);
		}
	});

	$page.find("button#game-end").on("vclick", function(event, ui) {
		cancelMoveChoice();

		if (g_editedDepth < g_editedGame.moves.length && !ABORT) {
			for (var i = g_editedDepth; i < g_editedGame.moves.length; i++)
				SP_PlayMove(VIEWER, -1);
			g_editedDepth = g_editedGame.moves.length;
			autoPlayPassMoves();
			updateBoard();
			updateSelectedMove();
		}
	});

	$page.on("vclick", "p#game-moves > span", function(event, ui) {
		if (ABORT)
			return;
		cancelMoveChoice();

		var depthStr = $(event.target).attr("class");
		var newDepth;
		if (depthStr == "initial")
			newDepth = 0;
		else {
			// Compute depth after move is played.
			newDepth = parseInt(depthStr) + 1;
			while (newDepth < g_editedGame.moves.length) {
				var internal = g_editedGame.moves[newDepth].internal;
				if (internal.substr(0, 5) != "-----")
					break;
				newDepth++;
			}
		}

		// Fix edited depth.
		while (g_editedDepth > newDepth) {
			SP_ReverseMove(VIEWER);
			g_editedDepth--;
		}
		while (g_editedDepth < newDepth) {
			SP_PlayMove(VIEWER, -1);
			g_editedDepth++;
		}

		// Need to autopass in case we pass after clicking "initial".
		autoPlayPassMoves();
		updateBoard();
		updateSelectedMove();
	});

/*

# Events if nothing changes:
touchstart vmousedown
touchend vmouseup vclick
(browser waits 300ms)
mousedown mouseup

# Events if the content changes underneath the point that was clicked:
touchstart vmousedown
touchend vmouseup vclick
(browser waits 300ms)
mousedown vmousedown mouseup vmouseup vclick

See https://api.jquerymobile.com/vclick/

	g_list = ["vclick", "vmousecancel", "vmousedown",
	          "vmouseup", "touchstart", "touchend", "mousedown", "mouseup"];
	for (var i = 0; i < g_list.length; i++) {
		$page.on(g_list[i], function(event, ui) {
			$("div#test-results").append(event.type + " " +
				(new Date()).toGMTString() + "<br>");
			$("div#test-results-box").show();
		});
	}
*/

	// Prevent body-scrolling on touch devices, making it possible to drag the
	// pieces. This should work on iOS, unlike the CSS alternative
	// "div.piece { touch-action: none; }".
	$page.on("touchmove", "td#game-board > div.piece", function(event, ui) {
		event.preventDefault();
	});

	$page.on("vmousedown", "td#game-board > div.piece", function(event, ui) {
		// This should prevent the browser from selecting text and from
		// changing the cursor to text. See stackoverflow 2745028.
		event.preventDefault();

		if (ABORT)
			return;
		cancelMoveChoice();

		g_mouseMoveSeen = false;
		if (g_moveState != 0)
			return;

		var $board = $("td#game-board");
		var offset = $board.offset();
		var size = pieceSize();
		g_mouseDownX = event.pageX - offset.left;
		g_mouseDownY = event.pageY - offset.top;
		var u = Math.floor(g_mouseDownX / size.x);
		var v = Math.floor(g_mouseDownY / size.y);

		// Remove old highlight.
//		$board.find("div.highlight").remove();

		g_originSquare = String.fromCharCode(97 + u, 56 - v);

		g_originPiece = $("td#game-board > div." + g_originSquare);
		if (g_originPiece.length == 0)
			return;

		// User just clicked on a piece.
		g_moveState = 1;

		$board.append("<div class='highlight' " +
		              "style='position:absolute; " +
		              "left:" + (u*size.x).toFixed(3) + "px; " +
		              "top:" + (v*size.y).toFixed(3) + "px; " +
		              "width:" + size.x.toFixed(3) + "px; " +
		              "height:" + size.y.toFixed(3) + "px;'></div>");

		// Create a list of possible destination squares.
		var destList = [];
		var initialStack = Module.stackSave();
		try {
			var legalMoves = Viewer_ComputeLegalMoves();
			var len = Viewer_GetMoveLength();
			for (var i = 0; i < legalMoves.length; i += len) {
				if (legalMoves.substr(i, 2) == g_originSquare) {
					var destSquare = legalMoves.substr(i+2, 2);
					var value = (1 << Viewer_HeuristicMoveType(i / len));
					addSquareToList(destList, destSquare, value);
				}
			}
		}
		catch (e) {
			catchFn(e);
			Module.stackRestore(initialStack);
		}

		// Add a marker for every destination square.
		for (var i = 0; i < destList.length; i++) {
			var square = destList[i][0];
			var value = destList[i][1];
			var color;
			if (value == 1)
				color = "green";
			else if (value == 2)
				color = "orange";
			else if (value == 4)
				color = "red";
			else
				color = "gray";
			var u1 = square.charCodeAt(0) - 97;
			var v1 = 56 - square.charCodeAt(1);
			$board.append("<div class='marker " + color + "' " +
			              "style='position:absolute; " +
			              "left:" + (u1*size.x).toFixed(3) + "px; " +
			              "top:" + (v1*size.y).toFixed(3) + "px; " +
			              "width:" + size.x.toFixed(3) + "px; " +
			              "height:" + size.y.toFixed(3) + "px;'></div>");
		}

		g_originPiece.css({
				"z-index": 1
			});
	});

	$page.on("vmousemove", function(event, ui) {
		if (g_moveState == 1) {
			var offset = $("td#game-board").offset();
			var x = event.pageX - offset.left;
			var y = event.pageY - offset.top;

			if (g_mouseMoveSeen ||
			    Math.abs(x - g_mouseDownX) >= 5 ||
			    Math.abs(y - g_mouseDownY) >= 5) {
				var size = pieceSize();

				g_originPiece.css({
						left: (x - size.x/2).toFixed(3) + "px",
						top: (y - size.y/2).toFixed(3) + "px",
						"z-index": 1
					});
				if (g_originPiece.hasClass("draggable")) {
					g_originPiece.removeClass("draggable");
					g_originPiece.addClass("dragging");
				}

				g_mouseMoveSeen = true;
			}
		}
	});

	$page.on("vmouseup", function(event, ui) {
		if (g_moveState == 0)
			return;

		if (g_moveState == 1 && !g_mouseMoveSeen) {
			// Mouseup was only selecting the origin square.
			g_moveState = 2;
			return;
		}

		var $board = $("td#game-board");
		var size = pieceSize();

		$board.find("div.highlight").remove();
		$board.find("div.marker").remove();

		// Put the piece back where we found it in case the move is illegal.
		// This doesn't matter if the move is legal.
		resetOriginPiece(size);

		g_moveState = 0;

		// This prevents the browser-generated vmousedown 300ms later
		// which wrongly reselects the piece.
		event.preventDefault();

		// Find destination square.
		var offset = $("td#game-board").offset();
		var x = event.pageX - offset.left;
		var y = event.pageY - offset.top;
		var u1 = Math.floor(x / size.x);
		var v1 = Math.floor(y / size.y);
		if (u1 >= 0 && u1 < 8 && v1 >= 0 && v1 < 8) {
			var destSquare = String.fromCharCode(97 + u1, 56 - v1);
			// Play move if legal.
			var matches = [];
			var legalMoves = Viewer_ComputeLegalMovesSafe();
			var len = Viewer_GetMoveLength();
			var old_k = -1;
			var internal = g_editedDepth < g_editedGame.moves.length ?
			               g_editedGame.moves[g_editedDepth].internal : "";
			for (var i = 0; i < legalMoves.length; i += len) {
				if (legalMoves.substr(i, 4) == g_originSquare + destSquare)
					matches.push(i / len);
				// Also reconstruct old k, if applicable.
				if (legalMoves.substr(i, len) == internal)
					old_k = i / len;
			}

			if (matches.length == 1)
				playMove(matches[0]);
			else if (matches.length > 1) {
				// Move the piece to the destination square, which looks better
				// when waiting for the choice. This is undone by
				// cancelMoveChoice.
				g_originPiece.css({
					left: (u1 * size.x).toFixed(3) + "px",
					top: (v1 * size.y).toFixed(3) + "px",
					"z-index": 0
				});

				// Add a select menu. First determine its position.
				// Heuristically try to have the button appear within
				// the board.
				var width = 2.75;
				var u = u1 + 0.5 - width/2;
				var v = v1 - 0.18;
				if (u < 0)
					u = 0;
				else if (u > 8 - width)
					u = 8 - width;

				var langObj = languageObject();
				var html = "<div style='position:absolute; " +
				           "left:" + (u*size.x).toFixed(3) + "px; " +
				           "top:" + (v*size.y).toFixed(3) + "px;'>" + 
				           "<select id='move-choice' data-inline='true'>\n";
				html += "<option value=''>Choose</option>\n";
				for (var i = 0; i < matches.length; i++) {
					var k = matches[i];
					html += "<option value='" + k + "'>";

					// Play the move to get its notation.
					SP_PlayMove(VIEWER, k);
					html += moveToText(getLastMove(VIEWER), langObj);
					SP_ReverseMove(VIEWER);

					html += "</option>\n";
				}
				if (old_k >= 0) {
					// We just messed up the variation stored in the engine's
					// viewer. Fix this because we sometimes call
					// SP_PlayMove(VIEWER, -1).
					SP_PlayMove(VIEWER, old_k);
					SP_ReverseMove(VIEWER);
				}
				html += "</select></div>\n";
				$board.append(html);

				// Initialize the select menu.
				$board.find("select").selectmenu();
			}
		}
	});

	$page.on("change", "select#move-choice", function(event, ui) {
		var k = parseInt($(event.target).val());
		playMove(k);
		$(event.target).remove();
	});

	$page.find("input#editable-game").on("change", function(event, ui) {
		cancelMoveChoice();

		var checked = $("input#editable-game").is(":checked");
		if (g_engineReady && !ABORT && !checked) {
			// Convert textarea to div.
			extractEditedGame();
			finalizeGameEditor();
			// TODO: erase old highlight only if something was changed.
			g_editedSpan.css("background-color", "");
		}

		applyEditableCheckbox();
	});

	$page.find("div#popupMenu").on("popupbeforeposition", function(event, ui) {
		// Adjust button names and disabled states.
		g_createStip1 = "dia";
		g_createStip2 = "a=>b";
		if (g_engineReady && !ABORT) {
			// Check if there is an initial position declaration.
			var detect = ["forsyth", "pieces"];
			for (var i = 0; i < g_commands.length; i++) {
				var command = g_commands[i];
				if (detect.indexOf(command.command) != -1) {
					g_createStip1 = "a=>b";
					break;
				}
			}

			var len = g_editedDepth;
			if (len > 0)
				g_createStip1 += " " + (len/2).toFixed(1);

			if (g_createdDepth != null) {
				var delta = g_editedDepth - g_createdDepth;
				if (delta > 0)
					g_createStip2 += " " + (delta/2).toFixed(1);
			}
		}

		var $button = $("button#create-new")
			.text("New " + g_createStip1 + " from start to current diagram");
		if (g_createStip1.indexOf(" ") == -1)
			$button.addClass("ui-disabled");
		else
			$button.removeClass("ui-disabled");

		$button = $("button#create-add")
			.text("Add " + g_createStip2 + " (b = current diagram)");
		if (g_createStip2.indexOf(" ") == -1)
			$button.addClass("ui-disabled");
		else
			$button.removeClass("ui-disabled");
	});

	$page.find("div#popupMenu").on("vclick", "button", function() {
 		if (!g_engineReady || ABORT)
			return;

		var id = $(this).attr("id");
		var langObj = languageObject(g_language);

		// Build diagram text.
		var squares = [];
		for (var i = 0; i < 64; i++)
			squares.push(Viewer_GetSquare(i));
		var diagramText = buildForsyth(squares, langObj);
		if (diagramText == "")	// No way to set royal in forsyth.
			diagramText = buildPieces(squares, langObj);
		diagramText += "\n";

		var stipKeyword = "stipulation"
		if (g_language == "french")
			stipKeyword = "enonce";
		else if (g_language == "german")
			stipKeyword = "forderung";

		var text;
		if (id == "create-new" ) {
			// Preserve some commands from problem.
			var commands = [];

			var keep = ["comment", "forsyth", "pieces", "set", "condition"];

			for (var i = 0; i < g_commands.length; i++) {
				var command = g_commands[i];
				if (keep.indexOf(command.command) != -1)
					commands.push(command);
			}
			var commandsText = textFromCommands(commands);

			// Make sure it doesn't start with a newline.
			if (commandsText.substr(0, 2) == "\r\n")
				commandsText = commandsText.substr(2);
			else if (commandsText[0] == "\n")
				commandsText = commandsText.substr(1);

			// Make sure it ends with a newline.
			if (commandsText != "" &&
			    commandsText[commandsText.length - 1] != "\n") {
				commandsText += "\n";
			}

			// Build language text if needed.
			var langText = "";
			if (g_language == "englishN")
				langText = "englishN\n";
			else if (g_language == "french")
				langText = "DebutProbleme\n";
			else if (g_language == "german")
				langText = "AnfangProblem\n";

			// Build option text if needed.
			var optionText = "";
			if (g_firstPly == 1)
				optionText = "option FirstToMove " + langObj.keywords[2] + "\n";
			if (g_initialMoveNumber != 1) {
				optionText += "option InitialMoveNumber " +
				              g_initialMoveNumber + "\n";
			}

			// Build stipulation text.
			var stipText = stipKeyword + " " + g_createStip1 + "\n";

			// Assemble new command text.
			text = langText + optionText + commandsText +
			       stipText + diagramText;
		}
		else if (id == "create-add") {
			// This should already end with a newline since we created it.
			var commandsText = textFromCommands(g_commands);

			// Build stipulation text.
			var stipText = stipKeyword + " " + g_createStip2 + "\n";

			// Assemble new command text.
			text = commandsText + stipText + diagramText;
		}
		else
			return;	// Something's wrong.

		$page.find("div#popupMenu").popup("close");
		$("textarea#text-representation").val(text)
		                                 .keyup();
		g_commands = [];
		$("button#apply-changes").trigger("vclick");

		// Do this last because apply-changes clears it.
		g_createdDepth = g_editedDepth;
	});

	// In case the checkbox is set after page reload.
	applyEditableCheckbox();
});
