//"use strict";

// We want this file to be pure ASCII to avoid character-encoding issues.
// We use a human-readable encoding of accents. Field meanings:
// index 0: RegExp matching the encoded version of a character
// index 1: the decoded character (unicode)
// index 2: German removal of accents (e.g. Zuerich)
// index 3: English/French removal of accents (e.g. Zurich)
var accentCodes = [
	// Put the longest matches first.
	[/''a/g, "\u00E4", "ae", "a"],
	[/''o/g, "\u00F6", "oe", "o"],
	[/''u/g, "\u00FC", "ue", "u"],
	[/'s/g, "\u00DF", "ss", "ss"],
	[/'e/g, "\u00E9", "e", "e"],
	[/`e/g, "\u00E8", "e", "e"],
	[/\^a/g, "\u00E2", "a", "a"],
	[/'E/g, "\u00C9", "E", "E"],
	[/`A/g, "\u00C0", "A", "A"]
];

function transformAccents(array) {
	// Build a second array with the transformed entries.
	var ret = [];
	for (var i = 0; i < array.length; i++) {
		var parts = array[i].split(": ");
		if (parts.length == 1)
			ret.push(array[i]);
		else {
			var first = parts[0];	// Make a copy because we modify parts[0].
			var last_part0 = "";
			for (var j = 1; j < 4; j++) {
				var part0 = first;
				for (var k = 0; k < accentCodes.length; k++)
					part0 = part0.replace(accentCodes[k][0], accentCodes[k][j]);
				if (part0 != parts[1] && part0 != last_part0) {
					parts[0] = part0;
					ret.push(parts.join(": "));
					last_part0 = part0;
				}
			}
		}
	}

	// Copy the answer to the parameter array.
	array.length = 0;
	for (var i = 0; i < ret.length; i++)
		array.push(ret[i]);
}

var commandNames = [
	// Language commands.
	"BeginProblem: englishS",
	"D'ebutProbl`eme: french",
	"AnfangProblem: german",
	"EnglishN: englishN",

	// Commands to ignore.
	"EndProblem: ignore", "FinProbl`eme: ignore", "EndeProblem: ignore",

	// Commands that are comments.
	"remark: comment", "remarque: comment", "bemerkung: comment",
	"author: comment", "auteur: comment", "autor: comment",
	"origin: comment", "source: comment", "quelle: comment",
	"title: comment", "titre: comment", "titel: comment",
	"protocol: comment", "protokoll: comment",

	// Commands that do something.
	"forsyth",
	"pieces",
		"pi`eces: pieces",
		"steine: pieces",
	"option",
	"play",
	"test",
	"set",
	"stipulation",
		"'enonc'e: stipulation",
		"forderung: stipulation",
	"condition",
		"bedingung: condition",
	"constraints",
		"contraintes: constraints",
		"einschr''ankungen: constraints"
];
transformAccents(commandNames);

// Make these language-dependent because "black" and "blanc" share a long
// prefix. [0] and [1] are the preferred white and black names. Others are
// alternatives.
var pieceColorNames = {
	englishN: ["white", "black"],
	englishS: ["white", "black"],
	french: ["blanc: white", "noir: black", "blancs: white", "noirs: black"],
	german: ["weiss: white", "schwarz: black", "wei\u00DF: white"]
};

var pieceAttributeNames = {
	englishN: ["royal"],
	englishS: ["royal"],
	french: ["royale: royal", "royales: royal"],
	german: ["koeniglich: royal", "k\u00F6niglich: royal"]
};

var optionNames = [
	"FirstToMove",
	"InitialMoveNumber"
];

// The numbering needs to be kept in sync with the C engine.
var playNames = [
	"help: 0",
	"defense: 1", "defence: 1", "defensive: 1",
	"series: 2",
	"pass: 3"
];

// The numbering needs to be kept in sync with the C engine.
// We can't have anything which starts with h, *, s or r.
var testNames = [
	"not",	// negates the test
	"dia: 0", "a=>b: 0", "pg: 0", "pj: 0", "bp: 0",	// diagram
	"+: 1",	// check
	"#: 2",	// mate
	"=: 3",	// stalemate
	"!=: 4", // auto-stalemate
	"==: 5", 	// double-stalemate
	"!+: 6", // auto-check
	"!#: 7", 	// auto-mate
	"##: 8", 	// double-mate
	"x: 9",	// capture
	"%: 10",	// win a piece
	"00: 11", "OO: 11", "-00: 11", "-OO: 11",	// castling
	"ep: 12", "-ep: 12",	// en passant
	"2+: 13", "++: 13",	// double check
	"3+: 14", "+++: 14",	// triple check
	"4+: 15", "++++: 15",	// quadruple check
	"5+: 16", "+++++: 16",	// fivefold check
	"6+: 17", "++++++: 17",	// sixfold check
	"7+: 18", "+++++++: 18",	// sevenfold check
	"8+: 19", "++++++++: 19",	// eightfold check
	"$: 20",	// >= 1 provisional or true mate
	"$$: 21",	// >= 2 provisional or true mates
	"$$$: 22",	// >= 3 provisional or true mates
	"$$$$: 23",	// >= 4 provisional or true mates
	"$$$$$: 24",	// >= 5 provisional or true mates
	"~: 25"	// any move
//	"%",	// win-a-piece
//	"#=",	// mate or stalemate
//	http://bulletin.chessproblems.ca/pdf/cpb-02.pdf
//	"!=" // auto-stalemate (become stalemated)
//  "Z" // target-square
//  "ct" // "RK" circuit (a piece moves and come back to the square it occupied in the diagram position.)
//  "<>" // "PW" place-exchange (two pieces switch places)
//  "F" // pin (some non-king move is illegal because of self-check?)
//  "!F" // auto-pin (pins own unit)
//	http://www.jurajlorinc.com/chess/explanat.htm
//	"##"	// beidmatt: it is the position in which both kings are mated. The last move reaching this position may be played even if it leaves own king under check or it is selfcheck, but the king of the side playing the move reaching beidmatt mustn't be mated yet before this move is played
//	"##!"	// gegenmatt: beidmatt except that the king of the side playing the move reaching gegenmatt must be mated before that move. Gegenmatt is usually called doublemate in English, beidmatt having no counterpart. 

/*

begin
forsyth 8/8/R1K1k2r/8/8/1rP5/2r5/5R2
stip h##!1.0
end

  1.Rc2*c3 + Kc6-d6 ##!

  1.Rb3*c3 + Kc6-d6 ##

*/

];

// The numbering needs to be kept in sync with the C engine.
// TODO: use translated color names.

var setNames = [
	"test: 1",
	"pass: 2"
];

var setOptions = [
	// Options for "set test"
	"first: 1",
	"second: 2",
	"white: 3",
	"black: 4",

	"accept: 101",
	"reject: 102",
	"prefer: 103",
	"avoid: 104",
	"only: 105",
	"none: 106",
	"one: 107",	// Accept move only if it is the only one working.

	"wake",		// Wake idle side if special moves exist.
	"end",

	// Convenience options.
	"reflex",	// Shorthand for "prefer wake end".
	"living",	// Cancels the "end" option.

	// Option for "set pass".
	"unique",	// Accept parry move only if it is unique, similar
	         	// to "defense" processing.

	// Applies to both.
	"pure"		// Applies only to play type "series".
];

var prefixNames = [
	// fixed length with no closure
	"",		// for non-dia, must have length 0.5, equivalent to h0.5
	   		// for dia, equivalent to h.
	"s",	// must have length 1.0, equivalent to hs1.0
	"r",	// must have length 1.0, equivalent to hr1.0

	// help
	"h",
	"ph",
	"ah",
	"pah",
	"u",	// like h but starts with white, second player moves are unique
	// We could add pu, au, pau but no one will use them.

	// series
	"ser-",
	"phser-",
	"ahser-",
	"pahser-",
	"puser-",		// h replaced with u
	"auser-",
	"pauser-"
];

// Closure consists of any sequence of h and * and optionally end with s or r.

var diagramNames = [
	"forsyth",
	"pieces", "pi`eces: pieces", "steine: pieces"
];

var diagramOptions = [
	"Add",
		"Ajouter: Add",
		"Erg''anze: Add",
	"AddPieces: AddWhitePieces AddBlackPieces",
		"AjouterDesPi`eces: AddWhitePieces AddBlackPieces",
		"Erg''anzeSteine: AddWhitePieces AddBlackPieces",
	"AddWhitePieces",
		"AjouterDesPi`ecesBlanches: AddWhitePieces",
		"Erg''anzeWei'seSteine: AddWhitePieces",
	"AddBlackPieces",
		"AjouterDesPi`ecesNoires: AddBlackPieces",
		"Erg''anzeSchwarzeSteine: AddBlackPieces",
	"ColorThePieces",
		"ColorierLesPi`eces: ColorThePieces",
		"F''arbeDieSteine: ColorThePieces",
		"ColourThePieces: ColorThePieces"
];
transformAccents(diagramOptions);

var constraintsOptions = [
	"or",
		"ou: or",
		"oder: or"
];

var conditionNames = [
	"Chess960",
	"Chigorin",
	"Almost",
	"Knightmate",
	"CavalierMajeur",
	"Berolina",
	"NonRoyalKings",
	"NonRoyalKings_CastlingAllowed",
	"Extinction",
	"RoyalDynasty",
	"SiameseKings",
	"Vogtlaender",
	"LeseMajeste",

	"PowerTransfer",
	"PowerTransfer_TypeCirce",
	"PowerTransfer_RexInclusive",
	"PowerTransfer_PawnExclusive",
	"PowerTransfer_Mirror",
	"PowerTransfer_File",
	"PowerTransfer_Symmetry",
	"PowerTransfer_VerticalSymmetry",
	"PowerTransfer_HorizontalSymmetry",
	"PowerTransfer_Antipodes",
	"PowerTransfer_Above",
	"PowerTransfer_Below",
	"PowerTransfer_Diagram",
	"PowerTransfer_Diametral",
	"PowerTransfer_VerticalMirror",
	"PowerTransfer_FriendOnly",
	"PowerTransfer_EnemyOnly",

	"Bolero",
	"Bolero_RexInclusive",
	"Bolero_PawnInclusive",
	"Bolero_Inverse",

	"VerticalCylinder",
	"VerticalCylinder_NoExtendedCastling",
	"NormalPawn",

	"Monochromatic",
	"Monochromatic_Anti",
	"Grid",
	"WhiteEdgemover",
	"BlackEdgemover",
	"GrowingMen",
	"ShrinkingMen",

	"Madrasi",
	"Madrasi_RexInclusive",
	"Madrasi_PawnExclusive",
	"Eiffel",

	"Woozles",
	"Woozles_RexExclusive",
	"BiWoozles",
	"BiWoozles_RexExclusive",

	"Patrol",
	"Patrol_Anti",
	"Patrol_Ultra",
	"Provocation",
	"Provocation_Anti",
	"Provocation_Ultra",
	"Geneva",
	"Geneva_RexInclusive",
	"SuperGuards",
	"Immun",
	"Immun_RexInclusive",
	"Disparate",

	"MarsCirce",
	"MarsCirce_RexExclusive",
	"MarsCirce_PawnExclusive",
	"MarsCirce_Couscous",
	"MarsCirce_Mirror",
	"MarsCirce_File",
	"MarsCirce_Symmetry",
	"MarsCirce_VerticalSymmetry",
	"MarsCirce_HorizontalSymmetry",
	"MarsCirce_Antipodes",
	"MarsCirce_Above",
	"MarsCirce_Below",
	"MarsCirce_Diagram",
	"MarsCirce_Make",
	"MarsCirce_Diametral",
	"MarsCirce_VerticalMirror",
	"MarsCirce_Cheylan",

	"FreeForAll",
	"Bicapture",
	"Bicapture_RexExclusive",
	"Cannibal",

	"Multicapture",
	"Multicapture_Anti",

	"Kamikaze",
	"Kamikaze_RexInclusive",
	"Chameleon",
	"Chameleon_Reversal",
	"Relegation",
	"Einstein",
	"Einstein_Reversal",
	"Einstein_Anti",
	"Einstein_WithPromotions",

	"Glasgow",
	"PromOnly",
	"PromOnly_P",
	"PromOnly_S",
	"PromOnly_B",
	"PromOnly_R",
	"PromOnly_Q",
	"PromOnly_K",
	"AdversePromotions",
	"KobulKings",

	"AntiCirce",
	"AntiCirce_RexExclusive",
	"AntiCirce_PawnExclusive",
	"AntiCirce_Clone",
	"AntiCirce_Chameleon",
	"AntiCirce_Couscous",
	"AntiCirce_Mirror",
	"AntiCirce_File",
	"AntiCirce_Symmetry",
	"AntiCirce_VerticalSymmetry",
	"AntiCirce_HorizontalSymmetry",
	"AntiCirce_Antipodes",
	"AntiCirce_Ibidem",
	"AntiCirce_Above",
	"AntiCirce_Below",
	"AntiCirce_Diagram",
	"AntiCirce_Equipollents",
	"AntiCirce_Platzwechsel",
	"AntiCirce_Super",
	"AntiCirce_Cage",
	"AntiCirce_Make",
	"AntiCirce_Diametral",
	"AntiCirce_VerticalMirror",
	"AntiCirce_Assassin",
	"AntiCirce_Cheylan",
	"AntiCirce_MovedRebirth",
	"AntiCirce_BeforePromotion",
	"AntiCirce_OrthodoxChecks",
	"AntiCirce_EntireMove",
	"AntiCirce_Relaxed",

	"Breton",
	"Breton_Adverse",
	"Breton_Chromatic",
	"Breton_Py",

	"Tibet",
	"Tibet_Double",
	"Andernach",
	"AntiAndernach",
	"ChameleonPursuit",
	"Volage",
	"Volage_Hyper",

	"Circe",
	"Circe_RexInclusive",
	"Circe_PawnExclusive",
	"Circe_Clone",
	"Circe_Chameleon",
	"Circe_Couscous",
	"Circe_Mirror",
	"Circe_File",
	"Circe_Symmetry",
	"Circe_VerticalSymmetry",
	"Circe_HorizontalSymmetry",
	"Circe_Antipodes",
	"Circe_Ibidem",
	"Circe_Above",
	"Circe_Below",
	"Circe_Diagram",
	"Circe_Equipollents",
	"Circe_Platzwechsel",
	"Circe_Super",
	"Circe_Cage",
	"Circe_Make",
	"Circe_Diametral",
	"Circe_VerticalMirror",
	"Circe_Turncoats",
	"Circe_Assassin",
	"Circe_Strict",
	"Circe_MovedRebirth",
	"Circe_WaitMove",
	"Circe_WaitCapture",

	"Sentinels",
	"Sentinels_PionAdverse",
	"Sentinels_Berolina",
	"Sentinels_Para",
	"Sentinels_Totalmax16",

	"Contagion",
	"Contagion_ByWhite",
	"Contagion_ByBlack",
	"WhiteOscillatingKings",
	"WhiteOscillatingKings_TypeC",
	"BlackOscillatingKings",
	"BlackOscillatingKings_TypeC",
	"Masand",
	"Masand_Generalized",
	"#R",
	"#C",
	"Kk",
	"ReleasingCheck",

	"NoImmediateRepetition",
	"Koeko",
	"Koeko_Anti",
	"Koeko_New",
	"Koeko_AffectedByGrid",
	"Koeko_AppliesToRebirth",
	"Distancing",
	"Distancing_WhiteAvoidsBlack",
	"Distancing_WhiteAvoidsWhite",
	"Distancing_BlackAvoidsWhite",
	"Distancing_BlackAvoidsBlack",
	"Distancing_OrthodoxChecks",
	"Distancing_RexInclusive",
	"Isardam",
	"Isardam_TypeB",
	"Isardam_RexInclusive",
	"Isardam_PawnExclusive",
	"Leffie",
	"Leffie_TypeB",
	"RexMultiplex",
	"Checkless",
	"AllowTerminalAutoChecks",
	"AntiParry",
	"Rapid",
	"BackHome",

	"Duelist",
	"Duelist_Anti",
	"Duelist_Ultra",
	"Duelist_CastledRookInclusive",
	"Relay",
	"Relay_Ultra",
	"WhiteMustCapture",
	"WhiteMustCapture_Anti",
	"WhiteMustCapture_Ultra",
	"WhiteMustCapture_SemiUltra",
	"BlackMustCapture",
	"BlackMustCapture_Anti",
	"BlackMustCapture_Ultra",
	"BlackMustCapture_SemiUltra",
	"WhiteMustCheck",
	"WhiteMustCheck_Anti",
	"WhiteMustCheck_Ultra",
	"WhiteMustCheck_SemiUltra",
	"BlackMustCheck",
	"BlackMustCheck_Anti",
	"BlackMustCheck_Ultra",
	"BlackMustCheck_SemiUltra",
	"WhiteFollowMyLeader",
	"BlackFollowMyLeader",
	"WhiteAlphabetic",
	"WhiteAlphabetic_Anti",
	"BlackAlphabetic",
	"BlackAlphabetic_Anti",
	"WhiteMaximummer",
	"WhiteMaximummer_Anti",
	"BlackMaximummer",
	"BlackMaximummer_Anti",

	"BlackPassesIfStuck"
];

// These combinations deserve to be shown in the UI.
var comboNames = [
	"Annan:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_FriendOnly",
	"Nanna:" +
		" PowerTransfer PowerTransfer_Above PowerTransfer_RexInclusive" +
		" PowerTransfer_FriendOnly",
	"BackToBack:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_EnemyOnly",
	"FaceToFace:" +
		" PowerTransfer PowerTransfer_Above PowerTransfer_RexInclusive" +
		" PowerTransfer_EnemyOnly",
	"PointReflection:" +
		" PowerTransfer PowerTransfer_Symmetry PowerTransfer_RexInclusive",
	"Bichromatic: Monochromatic Monochromatic_Anti",
	"StagnatingMen: GrowingMen ShrinkingMen",
	"Lortap: Patrol Patrol_Anti",
	"Functionary: Provocation Provocation_Ultra",
	"AntiFunctionary: Provocation Provocation_Anti Provocation_Ultra",
	"Make&Take: MarsCirce MarsCirce_Make MarsCirce_Couscous",
	"Unicapture: Multicapture Multicapture_Anti",
	"Take&Make:" +
		" AntiCirce AntiCirce_Make AntiCirce_Couscous" +
		" AntiCirce_BeforePromotion AntiCirce_OrthodoxChecks" +
		" AntiCirce_MovedRebirth",
	"Rifle:" +
		" AntiCirce AntiCirce_Equipollents AntiCirce_Mirror" +
		" AntiCirce_BeforePromotion AntiCirce_MovedRebirth" +
		" AntiCirce_Relaxed",
	"SuperAndernach: Andernach AntiAndernach",
	"ParrainCirce: Circe Circe_WaitMove Circe_Equipollents",
	"ContraParrain: Circe Circe_WaitMove Circe_Equipollents Circe_Mirror",
	"PWC: Circe Circe_Platzwechsel",
	"CirceDoubleAgents: Circe Circe_Turncoats Circe_Mirror",
	"AntiTake&Make: Circe Circe_Make Circe_Strict Circe_MovedRebirth",
	"MemoryCirce: Circe Circe_WaitCapture Circe_Ibidem",
	"KamikazeCirce: AntiCirce AntiCirce_RexExclusive AntiCirce_Relaxed Circe",
	"SwappingKings:" +
		" WhiteOscillatingKings WhiteOscillatingKings_TypeC" +
		" BlackOscillatingKings BlackOscillatingKings_TypeC",
	"AllIn: FreeForAll NoImmediateRepetition",
	"ContactGrid: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Pandemic:" +
		" Contagion Contagion_ByBlack Distancing Distancing_WhiteAvoidsBlack",
	"FuddledMen: Duelist Duelist_Anti Duelist_Ultra",
	"Losing: NonRoyalKings WhiteMustCapture BlackMustCapture",
	"NoWhiteCapture:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra",
	"NoBlackCapture:" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"WhiteUltraSchachZwang: WhiteMustCheck WhiteMustCheck_SemiUltra",
	"BlackUltraSchachZwang: BlackMustCheck BlackMustCheck_SemiUltra",
	"WhiteMinimummer: WhiteMaximummer WhiteMaximummer_Anti",
	"BlackMinimummer: BlackMaximummer BlackMaximummer_Anti",
	"BlackCap:" +
		" BlackMustCapture BlackMustCapture_SemiUltra BlackPassesIfStuck",
	"BlackChecks:" +
		" BlackMustCheck BlackMustCheck_SemiUltra BlackPassesIfStuck",
	"Zigzag:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_SemiUltra" +
		" WhiteMustCheck WhiteMustCheck_Anti WhiteMustCheck_SemiUltra",
	"ZigzagStapff:" +
		" WhiteMustCapture WhiteMustCapture_Anti" +
		" WhiteMustCheck WhiteMustCheck_Anti",
	"ZigzagUltra:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" WhiteMustCheck WhiteMustCheck_Anti WhiteMustCheck_Ultra"
];

var otherNames = [
	// Popeye English compatibility.
	"MirrorCirce: Circe Circe_Mirror",
	"Hypervolage: Volage Volage_Hyper",
	"BichromChess: Monochromatic Monochromatic_Anti",
	"MonochromChess: Monochromatic",
	"GridChess: Grid",
	"K''oko: Koeko",
	"NoCapture:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"ImmunChess: Immun",
	"ImmuneChess: Immun",
	"ContactGridChess: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"ChameleonCirce: Circe Circe_Chameleon",
	"CouscousCirce: Circe Circe_Couscous",
	"EquipollentsCirce: Circe Circe_Equipollents",
	"FileCirce: Circe Circe_File",
	"DoubleTibet: Tibet Tibet_Double",
	"DiagramCirce: Circe Circe_Diagram",
	"DuellistChess: Duelist",
	"SymmetryCirce: Circe Circe_Symmetry",
	"Vogtl''anderChess: Vogtlaender",
	"EinsteinChess: Einstein",
	"NewK''oko: Koeko Koeko_New",
	"CirceClone: Circe Circe_Clone",
	"MirrorFileCirce: Circe Circe_File Circe_Mirror",
	"AntipodeanCirce: Circe Circe_Antipodes",
	"MirrorCirceClone: Circe Circe_Clone Circe_Mirror",
	"MirrorAntiCirce: AntiCirce AntiCirce_Mirror",
	"DiagramAntiCirce: AntiCirce AntiCirce_Diagram",
	"FileAntiCirce: AntiCirce AntiCirce_File",
	"SymmetryAntiCirce: AntiCirce AntiCirce_Symmetry",
	"MirrorFileAntiCirce: AntiCirce AntiCirce_File AntiCirce_Mirror",
	"AntipodeanAntiCirce: AntiCirce AntiCirce_Antipodes",
	"EquipollentsAntiCirce: AntiCirce AntiCirce_Equipollents",
	"AntiCloneCirce: AntiCirce AntiCirce_Clone",
	"AntiSuperCirce: AntiCirce AntiCirce_Super",
	"Take&MakeChess:" +
		" AntiCirce AntiCirce_Make AntiCirce_Couscous" +
		" AntiCirce_BeforePromotion AntiCirce_OrthodoxChecks" +
		" AntiCirce_MovedRebirth",
	"ReversalEinsteinChess: Einstein Einstein_Reversal",
	"SuperCirce: Circe Circe_Super",
	"RelegationChess: Relegation",
	"AndernachChess: Andernach",
	"ChameleonChess: Chameleon",
	"FunctionaryChess: Provocation Provocation_Ultra",
	"GlasgowChess: Glasgow",
	"AntiAndernachChess: AntiAndernach",
	"VerticalMirrorCirce: Circe Circe_VerticalMirror",
	"ChecklessChess: Checkless",
	"DiametralCirce: Circe Circe_Diametral",
	"MarsMirrorCirce: MarsCirce MarsCirce_Mirror",
	"AntiEinsteinChess: Einstein Einstein_Anti",
	"MirrorCouscousCirce: Circe Circe_Couscous Circe_Mirror",
	"CirceAssassin: Circe Circe_Assassin",
	"ExtinctionChess: Extinction",
	"EiffelChess: Eiffel",
	"AlphabeticChess: WhiteAlphabetic BlackAlphabetic",
	"CirceTurncoats: Circe Circe_Turncoats",
	"UltraPatrol: Patrol Patrol_Ultra",
	"AnnanChess:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_FriendOnly",
	"GenevaChess: Geneva",
	"AntiK''oko: Koeko Koeko_Anti",
	"ProvocationChess: Provocation",
	"CageCirce: Circe Circe_Cage",
	"WhiteAlphabeticChess: WhiteAlphabetic",
	"BlackAlphabeticChess: BlackAlphabetic",
	"CirceTake&Make: Circe Circe_Make Circe_Strict",
	"AntiCloneCirce: AntiCirce AntiCirce_Clone",
	"VerticalSymmetryCirce: Circe Circe_VerticalSymmetry",
	"HorizontalSymmetryCirce: Circe Circe_HorizontalSymmetry",
	"VerticalSymmetryAntiCirce: AntiCirce AntiCirce_VerticalSymmetry",
	"HorizontalSymmetryAntiCirce: AntiCirce AntiCirce_HorizontalSymmetry",
	"Make&TakeChess: MarsCirce MarsCirce_Make MarsCirce_Couscous",
//	"AntiCirce_PWC: AntiCirce_Platzwechsel",	// Not sure we want this one.

	"Circe_LastMove: Circe_WaitMove",
	"Circe_Parrain: Circe_WaitMove Circe_Equipollents",
	"Circe_ContraParrain: Circe_WaitMove Circe_Equipollents Circe_Mirror",
	"Circe_PWC: Circe_Platzwechsel",

	"Sentinels_PionAdvers: Sentinels_PionAdverse",
	"Sentinels_ParaSentinelles: Sentinels_Para",

	// Popeye French compatibility.
	"Circ'e: Circe",
	"Circ'eMal'efique: Circe Circe_Mirror",
	"Bichromatique: Monochromatic Monochromatic_Anti",
	"Monochromatique: Monochromatic",
	"Grille: Grid",
	"NoirBordCoup: BlackEdgemover",
	"BlancBordCoup: WhiteEdgemover",
	"Patrouille: Patrol",
	"Circe'Echange: Circe Circe_Platzwechsel",
	"SansPrises:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"ContactGrille: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Circ'eCam'el'eon: Circe Circe_Chameleon",
	"Circ'eCouscous: Circe Circe_Couscous",
	"Circ'e'Equipollents: Circe Circe_Equipollents",
	"NoirMaximum: BlackMaximummer",
	"NoirMinimum: BlackMaximummer BlackMaximummer_Anti",
	"BlancMaximum: WhiteMaximummer",
	"BlancMinimum: WhiteMaximummer WhiteMaximummer_Anti",
	"Sentinelles: Sentinels",
	"Circ'eDiagramme: Circe Circe_Diagram",
	"NoirPriseForce: BlackMustCapture",
	"BlancPriseForce: WhiteMustCapture",
	"NoirPoursuite: BlackFollowMyLeader",
	"BlancPoursuite: WhiteFollowMyLeader",
	"Duelliste: Duelist",
	"Circ'eParrain: Circe Circe_WaitMove Circe_Equipollents",
	"Circ'eSym'etrique: Circe Circe_Symmetry",
	"'EchecsVogtl''ander: Vogtlaender",
	"'EchecsEinstein: Einstein",
	"K''okoNouveaux: Koeko Koeko_New",
	"FileCirc'eMal'efique: Circe Circe_File Circe_Mirror",
	"Circ'eAntipoden: Circe Circe_Antipodes",
	"Circ'eCloneMal'efique: Circe Circe_Clone Circe_Mirror",
	"AntiCirc'eMal'efique: AntiCirce AntiCirce_Mirror",
	"AntiCirc'eDiagramme: AntiCirce AntiCirce_Diagram",
	"AntiFileCirc'e: AntiCirce AntiCirce_File",
	"AntiCirc'eSym'etrique: AntiCirce AntiCirce_Symmetry",
	"AntiFileCirc'eMal'efique: AntiCirce AntiCirce_File AntiCirce_Mirror",
	"AntiCirc'eAntipoden: AntiCirce AntiCirce_Antipodes",
	"AntiCirc'e'Equipollents: AntiCirce AntiCirce_Equipollents",
	"Revers'EchecsEinstein: Einstein Einstein_Reversal",
	"'EchecsAndernach: Andernach",
	"'EchecsCam'el'eon: Chameleon",
	"'EchecsFonctionnaire: Provocation Provocation_Ultra",
	"'EchecsGlasgow: Glasgow",
	"'EchecsAntiAndernach: AntiAndernach",
	"Circ'eMal'efiqueVerticale: Circe Circe_VerticalMirror",
	"Sans'Echecs: Checkless",
	"Circ'eDiam'etrale: Circe Circe_Diametral",
	"MarsCirc'eMal'efique: MarsCirce MarsCirce_Mirror",
	"'EchecsAntiEinstein: Einstein Einstein_Anti",
	"Circ'eCouscousMal'efique: Circe Circe_Couscous Circe_Mirror",
	"'EchecsExtinction: Extinction",
	"'EchecsEiffel: Eiffel",
	"NoirUltraSchachZwang: BlackMustCheck BlackMustCheck_SemiUltra",
	"BlancUltraSchachZwang: WhiteMustCheck WhiteMustCheck_SemiUltra",
	"BlancSansPrises:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra",
	"NoirSansPrises:" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"'EchecsAlphab'etiques: WhiteAlphabetic BlackAlphabetic",
	"BlancRoisOscillant: WhiteOscillatingKings",
	"NoirRoisOscillant: BlackOscillatingKings",
	"UltraPatrouille: Patrol Patrol_Ultra",
	"Rois'Echanges:" +
		" WhiteOscillatingKings WhiteOscillatingKings_TypeC" +
		" BlackOscillatingKings BlackOscillatingKings_TypeC",
	"DynastieRoyale: RoyalDynasty",
	"Take&Make'Echecs:" +
		" AntiCirce AntiCirce_Make AntiCirce_Couscous" +
		" AntiCirce_BeforePromotion AntiCirce_OrthodoxChecks" +
		" AntiCirce_MovedRebirth",
	"Noir'Echecs:" +
		" BlackMustCheck BlackMustCheck_SemiUltra BlackPassesIfStuck",
	"Annan'Echecs:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_FriendOnly",
	"PionNormale: NormalPawn",	// Typo in Popeye.
	"Elliuortap: Patrol Patrol_Anti",
	"'EchecsGen`eve: Geneva",
	"Cam'el'eonPoursuite: ChameleonPursuit",
	"QuiPerdGagne: NonRoyalKings WhiteMustCapture BlackMustCapture",
	"'EchecsProvacateurs: Provocation",	// Typo in Popeye.
	"Circ'eCage: Circe Circe_Cage",
	"RoisKobul: KobulKings",
	"Blanc'EchecsAlphab'etiques: WhiteAlphabetic",
	"Noir'EchecsAlphab'etiques: BlackAlphabetic",
	"Supergardes: SuperGuards",
	"Retour: BackHome",
	"Face`AFace:" +
		" PowerTransfer PowerTransfer_Above PowerTransfer_RexInclusive" +
		" PowerTransfer_EnemyOnly",
	"Dos`ADos:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_EnemyOnly",
	"Circ'eSym'etriqueVerticale: Circe Circe_VerticalSymmetry",
	"Circ'eSym'etriqueHorizontale: Circe Circe_HorizontalSymmetry",
	"AntiCirc'eSym'etriqueVerticale: AntiCirce AntiCirce_VerticalSymmetry",
	"AntiCirc'eSym'etriqueHorizontale: AntiCirce AntiCirce_HorizontalSymmetry",
	"L`eseMajest'e: LeseMajeste",
	"Make&TakeEchecs: MarsCirce MarsCirce_Make MarsCirce_Couscous",
	"R'eflexionCentrale:" +
		" PowerTransfer PowerTransfer_Symmetry PowerTransfer_RexInclusive",

	"MarsCirce_RexExclusif: MarsCirce_RexExclusive",
	"MarsCirce_Mal'efique: MarsCirce_Mirror",
	"MarsCirce_Mal'efiqueVerticale: MarsCirce_VerticalMirror",
	"MarsCirce_Colonne: MarsCirce_File",
	"MarsCirce_Sym'etrie: MarsCirce_Symmetry",
	"MarsCirce_VerticaleSym'etrique: MarsCirce_VerticalSymmetry",
	"MarsCirce_HorizontaleSym'etrique: MarsCirce_HorizontalSymmetry",
	"MarsCirce_Diagramme: MarsCirce_Diagram",

	"AntiCirce_RexExclusif: AntiCirce_RexExclusive",
	"AntiCirce_Mal'efique: AntiCirce_Mirror",
	"AntiCirce_Mal'efiqueVerticale: AntiCirce_VerticalMirror",
	"AntiCirce_Cam'el'eon: AntiCirce_Chameleon",
	"AntiCirce_Colonne: AntiCirce_File",
	"AntiCirce_Sym'etrie: AntiCirce_Symmetry",
	"AntiCirce_VerticaleSym'etrique: AntiCirce_VerticalSymmetry",
	"AntiCirce_HorizontaleSym'etrique: AntiCirce_HorizontalSymmetry",
	"AntiCirce_Diagramme: AntiCirce_Diagram",
	"AntiCirce_'Echange: AntiCirce_Platzwechsel",
	"AntiCirce_Rel^ach'e: AntiCirce_Relaxed",

	"Breton_Chromatique: Breton_Chromatic",

	"Circe_RexInclusif: Circe_RexInclusive",
	"Circe_Mal'efique: Circe_Mirror",
	"Circe_Mal'efiqueVerticale: Circe_VerticalMirror",
	"Circe_Cam'el'eon: Circe Circe_Chameleon",
	"Circe_CoupDernier: Circe_WaitMove",
	"Circe_Colonne: Circe_File",
	"Circe_Sym'etrie: Circe_Symmetry",
	"Circe_VerticaleSym'etrique: Circe_VerticalSymmetry",
	"Circe_HorizontaleSym'etrique: Circe_HorizontalSymmetry",
	"Circe_Diagramme: Circe_Diagram",
	"Circe_'Echange: Circe_Platzwechsel",
	"Circe_Stricte: Circe_Strict",

	"Madrasi_RexInclusif: Madrasi_RexInclusive",
	"Woozles_RexExclusif: Woozles_RexExclusive",
	"BiWoozles_RexExclusif: BiWoozles_RexExclusive",
	"Isardam_RexInclusif: Isardam_RexInclusive",
	"Geneva_RexInclusif: Geneva_RexInclusive",

	// Popeye German compatibility.
	"SpiegelCirce: Circe Circe_Mirror",
	"BichromesSchach: Monochromatic Monochromatic_Anti",
	"MonochromesSchach: Monochromatic",
	"Gitterschach: Grid",
	"SchwarzerRandz''uger: BlackEdgemover",
	"Wei'serRandz''uger: WhiteEdgemover",
	"PatrouilleSchach: Patrol",
	"PlatzwechselCirce: Circe Circe_Platzwechsel",
	"Ohneschlag:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"Immunschach: Immun",
	"KontaktGitter: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Cham''aleonCirce: Circe Circe_Chameleon",
	"SchwarzerL''angstz''uger: BlackMaximummer",
	"SchwarzerK''urzestz''uger: BlackMaximummer BlackMaximummer_Anti",
	"Wei'serL''angstz''uger: WhiteMaximummer",
	"Wei'serK''urzestz''uger: WhiteMaximummer WhiteMaximummer_Anti",
	"TibetSchach: Tibet",
	"DoppeltibetSchach: Tibet Tibet_Double",
	"DiagrammCirce: Circe Circe_Diagram",
	"SchwarzerSchlagzwang: BlackMustCapture",
	"Wei'serSchlagzwang: WhiteMustCapture",
	"SchwarzerVerfolgungsz''uger: BlackFollowMyLeader",
	"Wei'serVerfolgungsz''uger: WhiteFollowMyLeader",
	"Duellantenschach: Duelist",
	"SymmetrieCirce: Circe Circe_Symmetry",
	"Vogtl''anderSchach: Vogtlaender",
	"EinsteinSchach: Einstein",
	"NeuK''oko: Koeko Koeko_New",
	"SpiegelFileCirce: Circe Circe_File Circe_Mirror",
	"AntipodenCirce: Circe Circe_Antipodes",
	"SpiegelCirceclone: Circe Circe_Clone Circe_Mirror",
	"AntiSpiegelCirce: AntiCirce AntiCirce_Mirror",
	"AntiDiagrammCirce: AntiCirce AntiCirce_Diagram",
	"AntiFileCirce: AntiCirce AntiCirce_File",
	"AntiSymmetrieCirce: AntiCirce AntiCirce_Symmetry",
	"AntiSpiegelFileCirce: AntiCirce AntiCirce_File AntiCirce_Mirror",
	"AntiAntipodenCirce: AntiCirce AntiCirce_Antipodes",
	"AntiEquipollentsCirce: AntiCirce AntiCirce_Equipollents",
	"ReversesEinsteinSchach: Einstein Einstein_Reversal",
	"Degradierung: Relegation",
	"AndernachSchach: Andernach",
	"Cham''aleonSchach: Chameleon",
	"BeamtenSchach: Provocation Provocation_Ultra",
	"GlasgowSchach: Glasgow",
	"AntiAndernachSchach: AntiAndernach",
	"VertikalesSpiegelCirce: Circe Circe_VerticalMirror",
	"OhneSchach: Checkless",
	"MarsSpiegelCirce: MarsCirce MarsCirce_Mirror",
	"AntiEinsteinSchach: Einstein Einstein_Anti",
	"SpiegelCouscousCirce: Circe Circe_Couscous Circe_Mirror",
	"AssassinCirce: Circe Circe_Assassin",
	"AusrottungsSchach: Extinction",
	"EiffelSchach: Eiffel",
	"SchwarzerUltraSchachZwang: BlackMustCheck BlackMustCheck_SemiUltra",
	"Wei'serUltraSchachZwang: WhiteMustCheck WhiteMustCheck_SemiUltra",
	"Wei'serOhneSchlag:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra",
	"SchwarzerOhneSchlag:" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"AlphabetischesSchach: WhiteAlphabetic BlackAlphabetic",
	"TurncoatCirce: Circe Circe_Turncoats",
	"DoppelAgentenCirce: Circe Circe_Turncoats Circe_Mirror",
	"Wei'seOszillierendeK''onige: WhiteOscillatingKings",
	"SchwarzeOszillierendeK''onige: BlackOscillatingKings",
	"TauschK''onige:" +
		" WhiteOscillatingKings WhiteOscillatingKings_TypeC" +
		" BlackOscillatingKings BlackOscillatingKings_TypeC",
	"K''onigsDynastie: RoyalDynasty",
	"Take&MakeSchach:" +
		" AntiCirce AntiCirce_Make AntiCirce_Couscous" +
		" AntiCirce_BeforePromotion AntiCirce_OrthodoxChecks" +
		" AntiCirce_MovedRebirth",
	"SchwarzSch''acher:" +
		" BlackMustCheck BlackMustCheck_SemiUltra BlackPassesIfStuck",
	"Annanschach:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_FriendOnly",
	"NormalBauern: NormalPawn",
	"GenferSchach: Geneva",
	"Cham''aleonVerfolgung: ChameleonPursuit",
	"Schlagschach: NonRoyalKings WhiteMustCapture BlackMustCapture",
	"ProvokationSchach: Provocation",
	"K''afigCirce: Circe Circe_Cage",
	"KontraParrain: Circe Circe_WaitMove Circe_Equipollents Circe_Mirror",
	"KobulK''onige: KobulKings",
	"Wei'sesAlphabetischesSchach: WhiteAlphabetic",
	"SchwarzesAlphabetischesSchach: BlackAlphabetic",
	"SuperDeckungen: SuperGuards",
	"NachHause: BackHome",
	"AngesichtZuAngesicht:" +
		" PowerTransfer PowerTransfer_Above PowerTransfer_RexInclusive" +
		" PowerTransfer_EnemyOnly",
	"R''uckenAnR''ucken:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_EnemyOnly",
	"VertikaleSymmetrieCirce: Circe Circe_VerticalSymmetry",
	"HorizontaleSymmetrieCirce: Circe Circe_HorizontalSymmetry",
	"VertikaleAntiSymmetrieCirce: AntiCirce AntiCirce_VerticalSymmetry",
	"HorizontaleAntiSymmetrieCirce: AntiCirce AntiCirce_HorizontalSymmetry",
	"Make&TakeSchach: MarsCirce MarsCirce_Make MarsCirce_Couscous",
	"Punktspiegelung:" +
		" PowerTransfer PowerTransfer_Symmetry PowerTransfer_RexInclusive",

	"MarsCirce_RexExklusive: MarsCirce_RexExclusive",
	"MarsCirce_Spiegel: MarsCirce_Mirror",
	"MarsCirce_VertikalSpiegel: MarsCirce_VerticalMirror",
	"MarsCirce_Symmetrie: MarsCirce_Symmetry",
	"MarsCirce_VertikaleSymmetrie: MarsCirce_VerticalSymmetry",
	"MarsCirce_HorizontaleSymmetrie: MarsCirce_HorizontalSymmetry",
	"MarsCirce_Diagramm: MarsCirce_Diagram",
	"MarsCirce_Antipoden: MarsCirce_Antipodes",

	"AntiCirce_RexExklusive: AntiCirce_RexExclusive",
	"AntiCirce_Spiegel: AntiCirce_Mirror",
	"AntiCirce_VertikalSpiegel: AntiCirce_VerticalMirror",
	"AntiCirce_Cham''aleon: AntiCirce_Chameleon",
	"AntiCirce_K''afig: AntiCirce_Cage",
	"AntiCirce_Symmetrie: AntiCirce_Symmetry",
	"AntiCirce_VertikaleSymmetrie: AntiCirce_VerticalSymmetry",
	"AntiCirce_HorizontaleSymmetrie: AntiCirce_HorizontalSymmetry",
	"AntiCirce_Diagramm: AntiCirce_Diagram",
	"AntiCirce_Antipoden: AntiCirce_Antipodes",
	"AntiCirce_Locker: AntiCirce_Relaxed",

	"Breton_Advers: Breton_Adverse",
	"Breton_Chromatisch: Breton_Chromatic",

	"Circe_RexInklusive: Circe_RexInclusive",
	"Circe_Spiegel: Circe_Mirror",
	"Circe_VertikalSpiegel: Circe_VerticalMirror",
	"Circe_Cham''aleon: Circe_Chameleon",
	"Circe_Turncoat: Circe_Turncoats",
	"Circe_LetzterZug: Circe_WaitMove",
	"Circe_K''afig: Circe_Cage",
	"Circe_Symmetrie: Circe_Symmetry",
	"Circe_VertikaleSymmetrie: Circe_VerticalSymmetry",
	"Circe_HorizontaleSymmetrie: Circe_HorizontalSymmetry",
	"Circe_Diagramm: Circe_Diagram",
	"Circe_Antipoden: Circe_Antipodes",
	"Circe_Strikt: Circe_Strict",

	"Madrasi_RexInklusive: Madrasi_RexInclusive",
	"Woozles_RexExklusive: Woozles_RexExclusive",
	"BiWoozles_RexExklusive: BiWoozles_RexExclusive",
	"Isardam_RexInklusive: Isardam_RexInclusive",
	"Geneva_RexInklusive: Geneva_RexInclusive",
	"WhiteOscillatingKings_TypC: WhiteOscillatingKings_TypeC",
	"BlackOscillatingKings_TypC: BlackOscillatingKings_TypeC",

	// Popeye applies some of these to the "defending side", but we apply to
	// black as this is the common meaning in publications.
	"Edgemover: BlackEdgemover",
	"Randz''uger: BlackEdgemover",
	"Maximummer: BlackMaximummer",
	"Minimummer: BlackMaximummer BlackMaximummer_Anti",
	"Maximum: BlackMaximummer",	// Keep this otherwise Maximum is ambiguous.
	"Minimum: BlackMaximummer BlackMaximummer_Anti",
	"L''angstz''uger: BlackMaximummer",
	"K''urzestz''uger: BlackMaximummer BlackMaximummer_Anti",
	"Schlagzwang: BlackMustCapture",
	"SchachZwang: BlackMustCheck",
	"UltraSchachZwang: BlackMustCheck BlackMustCheck_SemiUltra",

	// Other English names.
	"ChigorinChess: Chigorin",
	"AlmostChess: Almost",
	"KnightmateChess: Knightmate",
	"NightriderChess: CavalierMajeur",
	"BerolinaChess: Berolina",
	"BerolinaPawns: Berolina",
	"SouthernChess:" +
		" PowerTransfer PowerTransfer_Below PowerTransfer_RexInclusive" +
		" PowerTransfer_FriendOnly",
	"NannaChess:" +
		" PowerTransfer PowerTransfer_Above PowerTransfer_RexInclusive" +
		" PowerTransfer_FriendOnly",
	"CircePowerTransfer: PowerTransfer PowerTransfer_TypeCirce",
	"MonochromeChess: Monochromatic",
	"BichromeChess: Monochromatic Monochromatic_Anti",
	"DoubleEdgemover: WhiteEdgemover BlackEdgemover",
	"MonoWoozles: Woozles",
	"AntiUltraPatrol: Patrol Patrol_Anti Patrol_Ultra",
	"AntiProvocationChess: Provocation Provocation_Anti",	// My invention.
	"AntiFunctionaryChess: Provocation Provocation_Anti Provocation_Ultra",
	"FreeForAllChess: FreeForAll",
	"MulticaptureChess: Multicapture",
	"UnicaptureChess: Multicapture Multicapture_Anti",
	"RifleChess:" +
		" AntiCirce AntiCirce_Equipollents AntiCirce_Mirror" +
		" AntiCirce_BeforePromotion AntiCirce_MovedRebirth" +
		" AntiCirce_Relaxed",
	"KamikazeChess: Kamikaze",
	"AntiCapture: Kamikaze Kamikaze_RexInclusive Circe Circe_Ibidem",
	"AutoCapture: Kamikaze Kamikaze_RexInclusive Circe Circe_Platzwechsel",
	"ReversalChameleon: Chameleon Chameleon_Reversal",
//	"NoPromotion: PromOnly",
	"AntiCageCirce: AntiCirce AntiCirce_Cage",
	"Volages: Volage",
	"Hypervolages: Volage Volage_Hyper",
	"InterchangeCirce: Circe Circe_Platzwechsel",
	"IbidemCirce: Circe Circe_Ibidem",
	"VMC: Circe Circe_VerticalMirror",
	"#RChess: #R",
	"#Removal: #R",
	"#CChess: #C",
	"#Color: #C",
	"#Colour: #C",
	"KkChess: Kk",
	"OscillatingKings: WhiteOscillatingKings BlackOscillatingKings",
	"AllInChess: FreeForAll NoImmediateRepetition",
	"AntiNewK''oko: Koeko Koeko_Anti Koeko_New",
	"PandemicWithOrthodoxChecks:" +
		" Contagion Contagion_ByBlack Distancing Distancing_WhiteAvoidsBlack" +
		" Distancing_OrthodoxChecks",
	"PandemicWithoutDistancing: Contagion Contagion_ByBlack",
	"PandemicReverseColors:" +
		" Contagion Contagion_ByWhite Distancing Distancing_BlackAvoidsWhite",
	"PandemicReverseColours:" +
		" Contagion Contagion_ByWhite Distancing Distancing_BlackAvoidsWhite",
	"WhiteAntiK''oko:" +
		" Distancing Distancing_WhiteAvoidsBlack Distancing_WhiteAvoidsWhite" +
		" Distancing_RexInclusive",
	"BlackAntiK''oko:" +
		" Distancing Distancing_BlackAvoidsWhite Distancing_BlackAvoidsBlack" +
		" Distancing_RexInclusive",
	"RapidChess: Rapid",
	"DuelistChess: Duelist",
	"SingleCombatChess: Duelist",
	"RelayChess: Relay",
	"LosingChess: NonRoyalKings WhiteMustCapture BlackMustCapture",
	"WhiteUltraMustCheck: WhiteMustCheck WhiteMustCheck_Ultra",
	"BlackUltraMustCheck: BlackMustCheck BlackMustCheck_Ultra",
	"FollowMyLeader: BlackFollowMyLeader",
	"AlphabetChess: WhiteAlphabetic BlackAlphabetic",
	"AlphabeticalChess: WhiteAlphabetic BlackAlphabetic",
	"ABC: WhiteAlphabetic BlackAlphabetic",
	"AntiAlphabetChess:" +
		" WhiteAlphabetic WhiteAlphabetic_Anti" +
		" BlackAlphabetic BlackAlphabetic_Anti",
	"AntiAlphabeticChess:" +
		" WhiteAlphabetic WhiteAlphabetic_Anti" +
		" BlackAlphabetic BlackAlphabetic_Anti",
	"AntiAlphabeticalChess:" +
		" WhiteAlphabetic WhiteAlphabetic_Anti" +
		" BlackAlphabetic BlackAlphabetic_Anti",
	"DoubleMaximummer: WhiteMaximummer BlackMaximummer",
	"DoubleMinimummer:" +
		" WhiteMaximummer WhiteMaximummer_Anti" +
		" BlackMaximummer BlackMaximummer_Anti",
	"KamikazeCirceRexInclusive:" +
		" AntiCirce AntiCirce_Relaxed Circe Circe_RexInclusive",
	"KamikazeCirceMirror:" +
		" AntiCirce AntiCirce_RexExclusive AntiCirce_Relaxed AntiCirce_Mirror" +
		" Circe Circe_Mirror",
	"Checking: BlackMustCheck BlackMustCheck_SemiUltra BlackPassesIfStuck",

	"Circe_Exchange: Circe_Platzwechsel",
	"Circe_Interchange: Circe_Platzwechsel",
	"Circe_Memory: Circe_WaitCapture Circe_Ibidem",

	"Masand_Generalised: Masand_Generalized",
	"Relay_TypeB: Relay_Ultra",

	// Other French names.
	"'Echecs960: Chess960",
	"B'erolina: Berolina",
	"PionsB'erolina: Berolina",
	"RoisSiamois: SiameseKings",
	"Bol'ero: Bolero",
	"CylindreVertical: VerticalCylinder",
	"AntiUltraPatrouille: Patrol Patrol_Anti Patrol_Ultra",
	"Provocateurs: Provocation",
	"Fonctionnaires: Provocation Provocation_Ultra",
	"AntiFonctionnaires: Provocation Provocation_Anti Provocation_Ultra",
	"Bicaptures: Bicapture",
	"Cannibales: Cannibal",
	"Multicaptures: Multicapture",
	"Unicaptures: Multicapture Multicapture_Anti",
	"Fusil:" +
		" AntiCirce AntiCirce_Equipollents AntiCirce_Mirror" +
		" AntiCirce_BeforePromotion AntiCirce_MovedRebirth" +
		" AntiCirce_Relaxed",
	"Cam'el'eon: Chameleon",
	"D'egradation: Relegation",
//	"SansPromotion: PromOnly",
	"PromotionsAdverses: AdversePromotions",
	"Circ'eM'emoire: Circe Circe_WaitCapture Circe_Ibidem",
	"Circ'eKamikaze:" +
		" AntiCirce AntiCirce_RexExclusive AntiCirce_Relaxed Circe",
	"Pand'emie:" +
		" Contagion Contagion_ByBlack Distancing Distancing_WhiteAvoidsBlack",
	"AntiParade: AntiParry",
	"Rapide: Rapid",
	"Poursuite: BlackFollowMyLeader",
	"Alphab'etiques: WhiteAlphabetic BlackAlphabetic",
	"AntiAlphab'etiques:" +
		" WhiteAlphabetic WhiteAlphabetic_Anti" +
		" BlackAlphabetic BlackAlphabetic_Anti",
	"MaximumBlanc: WhiteMaximummer",
	"KamikazeCirc'eRexInclusif:" +
		" AntiCirce AntiCirce_Relaxed Circe Circe_RexInclusive",
	"KamikazeCirc'eMal'efique:" +
		" AntiCirce AntiCirce_RexExclusive AntiCirce_Relaxed AntiCirce_Mirror" +
		" Circe Circe_Mirror",

	"Bolero_RexInclusif: Bolero_RexInclusive",
	"Bolero_PionInclusif: Bolero_PawnInclusive",
	"Bolero_PionInclusiv: Bolero_PawnInclusive",

	"Madrasi_PionExclusif: Madrasi_PawnExclusive",
	"Madrasi_PionExclusiv: Madrasi_PawnExclusive",
	"Bicapture_RexExclusif: Bicapture_RexExclusive",
	"Kamikaze_RexInclusif: Kamikaze_RexInclusive",
	"Einstein_AvecPromotions: Einstein_WithPromotions",

	"MarsCirce_PionExclusif: MarsCirce_PawnExclusive",
	"AntiCirce_PionExclusif: AntiCirce_PawnExclusive",
	"Circe_PionExclusif: Circe_PawnExclusive",
	"Circe_M'emoire: Circe_WaitCapture Circe_Ibidem",
	"Masand_G'en'eralis'e: Masand_Generalized",

	// Other German names.
	"Schach960: Chess960",
	"Berolinaschach: Berolina",
	"Berolinabauern: Berolina",
	"SiamesischeK''onige: SiameseKings",
	"VertikalerZylinder: VerticalCylinder",
	"Majest''atsbeleidigung: LeseMajeste",
	"DoppelRandz''uger: WhiteEdgemover BlackEdgemover",
	"Ultrapatrouilleschach: Patrol Patrol_Ultra",
	"Provokateure: Provocation",
	"AntiBeamtenschach: Provocation Provocation_Anti Provocation_Ultra",
	"Kannibalenschach: Cannibal",
	"Flintenschach:" +
		" AntiCirce AntiCirce_Equipollents AntiCirce_Mirror" +
		" AntiCirce_BeforePromotion AntiCirce_MovedRebirth" +
		" AntiCirce_Relaxed",
	"Kamikazeschach: Kamikaze",
//	"OhneUW: PromOnly",
	"AntiK''afigCirce: AntiCirce AntiCirce_Cage",
	"Ged''achtniscirce: Circe Circe_WaitCapture Circe_Ibidem",
	"Antineuk''oko: Koeko Koeko_Anti Koeko_New",
	"Giko: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Gitterkontaktschach:" +
		" Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Schnellschach: Rapid",
	"Stafettenschach: Relay",
	"Verfolgungsz''uger: BlackFollowMyLeader",
	"Doppelverfolgungsz''uger: WhiteFollowMyLeader BlackFollowMyLeader",
	"Alphabetschach: WhiteAlphabetic BlackAlphabetic",
	"Doppell''angstz''uger: WhiteMaximummer BlackMaximummer",
	"Doppelk''urzestz''uger:" +
		" WhiteMaximummer WhiteMaximummer_Anti" +
		" BlackMaximummer BlackMaximummer_Anti",
	"KamikazeCirceRexInklusive:" +
		" AntiCirce AntiCirce_Relaxed Circe Circe_RexInclusive",
	"KamikazeSpiegelCirce:" +
		" AntiCirce AntiCirce_RexExclusive AntiCirce_Relaxed AntiCirce_Mirror" +
		" Circe Circe_Mirror",
	"UltraSchlagzwang: BlackMustCapture BlackMustCapture_Ultra",
	"Wei'serUltraSchlagzwang: WhiteMustCapture WhiteMustCapture_Ultra",
	"DoppelUltraSchlagzwang:" +
		" WhiteMustCapture WhiteMustCapture_Ultra" +
		" BlackMustCapture BlackMustCapture_Ultra",
	"WirklicherSchlagzwang: BlackMustCapture BlackMustCapture_SemiUltra",
	"Schwarzschl''ager:" +
		" BlackMustCapture BlackMustCapture_SemiUltra BlackPassesIfStuck",
	"Schlagzickzack:" +
		" BlackMustCapture BlackMustCapture_SemiUltra BlackPassesIfStuck" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_SemiUltra" +
		" WhiteMustCheck WhiteMustCheck_Anti WhiteMustCheck_SemiUltra",
	"Schachzickzack:" +
		" BlackMustCheck BlackMustCheck_SemiUltra BlackPassesIfStuck" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_SemiUltra" +
		" WhiteMustCheck WhiteMustCheck_Anti WhiteMustCheck_SemiUltra",
	"SchachzickzackTypStapff:" +
		" BlackMustCheck BlackMustCheck_SemiUltra BlackPassesIfStuck" +
		" WhiteMustCapture WhiteMustCapture_Anti" +
		" WhiteMustCheck WhiteMustCheck_Anti",
	"SchachzickzackUltraform:" +
		" BlackMustCheck BlackMustCheck_SemiUltra BlackPassesIfStuck" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" WhiteMustCheck WhiteMustCheck_Anti WhiteMustCheck_Ultra",

	"Bolero_RexInklusive: Bolero_RexInclusive",
	"Bolero_BauernInklusive: Bolero_PawnInclusive",
	"Bolero_BauernInclusiv: Bolero_PawnInclusive",

	"Madrasi_BauernExklusive: Madrasi_PawnExclusive",
	"Madrasi_BauernExclusiv: Madrasi_PawnExclusive",
	"Bicapture_RexExklusive: Bicapture_RexExclusive",
	"Kamikaze_RexInklusive: Kamikaze_RexInclusive",
	"Einstein_MitUmwandlungen: Einstein_WithPromotions",

	"MarsCirce_BauernExklusive: MarsCirce_PawnExclusive",
	"AntiCirce_BauernExklusive: AntiCirce_PawnExclusive",
	"Circe_BauernExklusive: Circe_PawnExclusive",
	"Circe_Ged''achtnis: Circe_WaitCapture Circe_Ibidem",

	"Relay_TypB: Relay_Ultra"
];
transformAccents(otherNames);

// testConditionList(comboNames);
// testConditionList(otherNames);
function testConditionList(list) {
	for (var i = 0; i < list.length; i++) {
		var parts = list[i].split(": ");
		if (parts.length != 2)
			console.log("Missing colon-space: " + list[i]);
		var parts2 = parts[1].split(" ");
		for (var j = 0; j < parts2.length; j++) {
			if (conditionNames.indexOf(parts2[j]) == -1)
				console.log("Not found: " + parts[0] + ": " + parts2[j]);
		}
	}
}

function isWhiteSpace(c) {
	return (c == " " || c == "\t" || c == "\n" || c == "\r" ||
	        c == "\u00A0");	// No-break space.
	// Maybe more characters should be matched. See
	// https://stackoverflow.com/questions/6073637/javascript-regex-whitespace-characters
}

// Searches "token" in "list", returning value, or one of these special values:
// "unknown", "ambiguous".
function findInList(token, list) {
	// First pass to extract everthing that matches.
	var matches = [];
	for (var i = 0; i < list.length; i++) {
		var parts = list[i].split(": ");
		var listToken = parts[0];
		var listValue = parts[parts.length - 1];

		if (listToken.length >= token.length &&
		    listToken.substr(0, token.length).toLowerCase() ==
			token.toLowerCase()) {
			matches.push([listToken, listValue]);
		}
	}
	if (matches.length == 0)
		return "unknown";

	// Circ will match Circe and CirceClone. Reject CirceClone because it is
	// a superstring of Circe.
	for (var i = 0; i < matches.length; i++) {
		if (matches[i][0] == "")	// Code for invalid.
			continue;
		// Remove listTokens that are superstrings of matches[i][0].
		for (var j = 0; j < matches.length; j++) {
			if (j == i)
				continue;
			if (matches[j][0].length > matches[i][0].length &&
			    matches[j][0].substr(0, matches[i][0].length) ==
			    matches[i][0]) {
				matches[j][0] = "";	// Code for invalid.
			}
		}
	}

	// Third pass to find matchValue.
	var matchValue = "";
	for (var i = 0; i < matches.length; i++) {
		if (matches[i][0] == "")
			continue;
		if (matchValue == "")
			matchValue = matches[i][1];
		else if (matches[i][1] != matchValue) {
			if (DEBUG)
				console.log(matches);
			return "ambiguous";
		}
	}

//	console.log(token + " in " + list + " found " + matchValue);
	return matchValue;
}

// Returns null if there is no match.
function grabLongestMatch(token, list) {
	var match = null;
	for (var i = 0; i < list.length; i++) {
		var parts = list[i].split(": ");
		if (token.substr(0, parts[0].length).toLowerCase() == parts[0] &&
		    (match == null || parts[0].length > match.length)) {
			match = parts[0];
		}
	}
	return match;
}

function grabClosure(token) {
	var i = 0;
	var c = token.charAt(i).toLowerCase();

	while (true) {
		if (c == "h" || c == "*") {
			i++;
			c = token.charAt(i);
		}
		else if (c == "s" || c == "r") {
			i++;
			break;
		}
		else
			break;
	}

	return token.substr(0, i);
}

// "name" is only used for the error message.
function readType(parser, command, name, list) {
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		openDialog("Error", name + " type is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var type = findInList(token, list);
	if (type == "unknown" || type == "ambiguous") {
		openDialog("Error", name + " type '" + token + "' is " + type + ".");
		return false;
	}

	command.type = {text: separator + token,
					value: type};
	parser.curr++;
	return true;
}

// Process a piece token such as "S". Returns false if the
// token isn't of that form.
function ParsePiece(token, language, color, addList) {
	var pieceLetter = token.toUpperCase();

	var piece = pieceLetters[language].indexOf(pieceLetter);
	if (piece < 0)
		return false;

	if (color == "white")
		piece += WHITE;
	else if (color == "black")
		piece += BLACK;
	else {
		openDialog("Error", "Unknown color '" + color + "'.");
		return false;
	}

	addList.push(piece);
	return true;
}

// Parse an AddList such as "white Q B R black P P".
function parseAddList(parser, command) {
	var addList = [];

	var text = "";

	var token = parser.getCurrToken();

	while (true) {
		if (token == "")
			break;	// Maybe we're done.
		var separator = parser.getCurrSeparator();

		var color = findInList(token, pieceColorNames[parser.language]);
		if (color == "ambiguous") {
			openDialog("Error", "Color name '" + token + "' is " + color + ".");
			return false;
		}
		else if (color == "unknown")
			break;	// Maybe we're done and token is the next command.

		// Color token is accepted!
		text += separator + token;

		// Parse pieces until the token isn't a piece anymore.
		while (true) {
			parser.curr++;

			token = parser.getCurrToken();
			if (token == "")
				break;
			if (!ParsePiece(token, parser.language, color, addList))
				break;
			separator = parser.getCurrSeparator();

			text += separator + token;
		}
	}

	if (addList.length > 32) {
		openDialog("Error", "Cannot add more than 32 pieces.");
		return false;
	}

	command.addList = {text: text, value: addList};
	return true;
}

// "name" is only used for the error message.
function readAnyOptions(parser, command, name, list) {
	command.options = [];

	while (true) {
		var token = parser.getCurrToken();
		if (token == "")		// Undefined or empty.
			break;
		var separator = parser.getCurrSeparator();

		var option = findInList(token, list);
		if (option == "ambiguous") {
			openDialog("Error", name + " option '" + token + "' is " +
			           option + ".");
			return false;
		}
		else if (option == "unknown")
			break;

		command.options.push({text: separator + token,
			                  value: option});
		parser.curr++;

		// Special diagram options.
		if (option == "Add") {
			if ("addList" in command) {
				openDialog("Error", name + " option '" + token +
				           "' cannot be repeated.");
				return false;
			}
			if (!parseAddList(parser, command))
				return false;
		}
	}
	return true;
}

// Parser functions.

function parseComment(parser, command) {
	while (parser.curr < parser.parts.length) {
		var separator = parser.getCurrSeparator();
		if (separator.search("[\n\r]") >= 0)
			break;
		var token = parser.getCurrToken();
		command.text += separator + token;
		parser.curr++;
	}
}

function parseForsyth(parser, command) {
	// Parse board.
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		openDialog("Error", "Forsyth string is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var T = pieceLetters[parser.language];

	var board = [];
	for (var i = 0; i < token.length; i++) {
		// Grab a number if there is one.
		var n = 0;
		while (true) {
			var d = token.charCodeAt(i);
			if (d >= 48 && d < 58) {
				n = n * 10 + (d - 48);
				i++;
				if (i < token.length)
					continue;
			}
			break;
		}
		if (board.length + n > 64) {
			openDialog("Error",
			           "Forsyth number '" + n + "' exceeds 64 squares.");
			return false;
		}
		while (n > 0) {
			board.push(EMPTY);
			n--;
		}

		if (i == token.length)
			break;

		var c = token.charAt(i);
		if (c == "/") {
			if (board.length == 0 || board.length == 64 ||
			    board.length % 8 != 0) {
				openDialog("Error",
				           "Forsyth row separator '/' is badly placed.");
				return false;
			}
			continue;
		}

		// Grab a fairy piece if there is one.
		if (c == ".") {
			var c2 = token.substr(i + 1, 2);
			if (c2.length != 2) {
				openDialog("Error",
				           "Forsyth dot '.' must be followed by 2 characters.");
				return false;
			}
			var j;
			for (j = 0; j < T.length; j++) {
				if (c2 == T[j]) {
					board.push(WHITE + j);
					break;
				}
				else if (c2 == T[j].toLowerCase()) {
					board.push(BLACK + j);
					break;
				}
			}
			if (j == T.length) {
				openDialog("Error",
				           "Forsyth fairy piece '." + c2 + "' is unknown.");
				return false;
			}
			i += 2;
			continue;
		}

		// Grab a piece letter.
		var j;
		for (j = 0; j < T.length; j++) {
			if (c == T[j]) {
				board.push(WHITE + j);
				break;
			}
			else if (c == T[j].toLowerCase()) {
				board.push(BLACK + j);
				break;
			}
		}
		if (j == T.length) {
			openDialog("Error", "Forsyth piece letter '" + c + "' is unknown.");
			return false;
		}
		if (board.length > 64) {
			openDialog("Error",
			           "Forsyth piece letter '" + c + "' exceeds 64 squares.");
			return false;
		}
	}

	if (board.length < 64) {
		openDialog("Error", "Forsyth string fills only " + board.length +
		           " squares.");
		return false;
	}

	command.position = {text: separator + token,
	                    value: {board: board}};
	parser.curr++;

	return true;
}

function squareFromString(str) {
	var file = str.charCodeAt(0);
	var rank = str.charCodeAt(1);
	if (file == 231) file = 99;	// c
	if (file == 233) file = 101;	// e
	if (file < 96) file += 32;	// toLowerCase
	file -= 97;	// a
	rank -= 49;	// 1
	if (file < 0 || file >= 8 || rank < 0 || rank >= 8)
		return -1;
	return (7-rank) * 8 + file;
}

// Process a piece-and-squares token such as "Sb1g1". Returns false if the
// token isn't of that form.
function ParsePieceAndSquares(token, language, color, royal_flag, position) {
	if (token.length < 3)
		return false;

	var pieceLetter = token.substr(0, (token.length + 1) % 2 + 1).toUpperCase();

	var piece = pieceLetters[language].indexOf(pieceLetter);
	if (piece < 0)
		return false;

	if (color == "white")
		piece += WHITE;
	else if (color == "black")
		piece += BLACK;
	else {
		openDialog("Error", "Unknown color '" + color + "'.");
		return false;
	}

	if (royal_flag)
		piece += ROYAL;

	var squares = [];
	for (var i = pieceLetter.length; i < token.length; i += 2) {
		var square = squareFromString(token.substr(i, 2));
		if (square >= 0)
			squares.push(square);
		else
			return false;
	}

	// Everything is good, add pieces to the board.
	for (var i = 0; i < squares.length; i++)
		position.board[squares[i]] = piece;

	return true;
}

function parsePieces(parser, command) {
	var position = {
		board: [],
		// We can add more fields later.
	};

	for (var i = 0; i < 64; i++)
		position.board[i] = EMPTY;

	var text = "";

	while (true) {
		var token = parser.getCurrToken();
		if (token == "")
			break;	// Maybe we're done.

		var color = findInList(token, pieceColorNames[parser.language]);
		if (color == "ambiguous") {
			openDialog("Error", "Color name '" + token + "' is " + color + ".");
			return false;
		}
		else if (color == "unknown")
			break;	// Maybe we're done and token is the next command.

		// Color token is accepted!
		var separator = parser.getCurrSeparator();
		text += separator + token;
		parser.curr++;

		// Look for attributes.
		var royal_flag = false;
		while (true) {
			token = parser.getCurrToken();
			if (token == "")
				break;	// Maybe we're done.
			var attribute = findInList(token,
			                           pieceAttributeNames[parser.language]);
			if (attribute == "ambiguous") {
				openDialog("Error", "Attribute name '" + token + "' is " +
				           attribute + ".");
				return false;
			}
			else if (attribute == "unknown")
				break;	// Done with attributes.

			// Attribute token accepted!
			var separator = parser.getCurrSeparator();
			text += separator + token;
			parser.curr++;

			if (attribute == "royal")
				royal_flag = true;
		}

		// Parse pieces until the token isn't a piece anymore.
		while (true) {
			token = parser.getCurrToken();
			if (!ParsePieceAndSquares(token, parser.language, color, royal_flag,
			                          position)) {
				break;
			}

			separator = parser.getCurrSeparator();
			text += separator + token;
			parser.curr++;
		}
	}

	command.position = {text: text, value: position};
	return true;
}

function parseOption(parser, command) {
	if (!readType(parser, command, "option", optionNames))
		return false;

	var type = command.type.value;
	if (type == "FirstToMove") {
		// Parse color.
		var token = parser.getCurrToken();
		if (token == "") {		// Undefined or empty.
			openDialog("Error", "FirstToMove color is missing.");
			return false;
		}
		var separator = parser.getCurrSeparator();

		var color = findInList(token, pieceColorNames[parser.language]);
		if (color == "unknown" || color == "ambiguous") {
			openDialog("Error", "Color name '" + token + "' is " + color + ".");
			return false;
		}

		command.color = {text: separator + token,
						 value: color};
		parser.curr++;
	}
	else if (type == "InitialMoveNumber") {
		// Parse number.
		var token = parser.getCurrToken();
		if (token == "") {		// Undefined or empty.
			openDialog("Error", "InitialMoveNumber value is missing.");
			return false;
		}
		var separator = parser.getCurrSeparator();

		var re = /^-?\d*(|\.0|\.5)$/;
		if (!re.test(token)) {
			openDialog("Error", "InitialMoveNumber '" + token + "' is wrong.");
			return false;
		}

		var number = parseFloat(token);

		command.number = {text: separator + token,
						  value: number};
		parser.curr++;
	}

	return true;
}

function parsePlay(parser, command) {
	if (!readType(parser, command, "play", playNames))
		return false;

	command.type.value = parseInt(command.type.value);

	// Parse length.
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		openDialog("Error", "Play length is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var re = /^\d*(|\.0|\.5)$/;
	if (!re.test(token)) {
		openDialog("Error", "Play length '" + token + "' is wrong.");
		return false;
	}
	parser.curr++;

	var length = parseFloat(token) * 2;

	// If a series, then reduce integer lengths by 1 ply.
	var series = (command.type.value ==
	              parseInt(findInList("series", playNames)));
	if (series) {
		re = /^\d*$/;
		if (re.test(token))
			length -= 1;
	}

	command.length = {text: separator + token,
	                  value: length};

	return true;
}

// "name" is typically "dia", "a=>b", or "PG" to make the error message less
// confusing.
function parseDia(parser, command, name) {
	parser.seenDia = true;

	// We require a diagram.
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		openDialog("Error", name + " diagram is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var type = findInList(token, diagramNames);
	command.dia = {text: separator + token,
		           command: type};
	parser.curr++;

	if (type == "forsyth") {
		if (!parseForsyth(parser, command.dia))
			return false;
	}
	else if (type == "pieces") {
		if (!parsePieces(parser, command.dia))
			return false;
	}
	else {
		openDialog("Error", name +
		           " must be immediately followed by forsyth or pieces.");
		return false;
	}

	return readAnyOptions(parser, command.dia, name, diagramOptions);
}

function parseTest(parser, command) {
	// Multiple passes in case there is a not.
	var aim;
	for (var i = 0; ; i++) {
		// Parse aim.
		var token = parser.getCurrToken();
		if (token == "") {		// Undefined or empty.
			openDialog("Error", "Aim is missing.");
			return false;
		}
		var separator = parser.getCurrSeparator();

		aim = findInList(token, testNames);

		if (i == 0 && aim == "not") {
			command.negated = {text: separator + token,
		                	   value: true};
			parser.curr++;
		}
		else
			break;
	}

	if (aim == "unknown" || aim == "ambiguous" || aim == "not") {
		openDialog("Error", "Aim '" + token + "' is " + aim + " .");
		return false;
	}

	command.aim = {text: separator + token,
				   value: parseInt(aim)};
	parser.curr++;

	if (aim == "0")	// dia.
		return parseDia(parser, command, token);

	return true;
}

function parseSet(parser, command) {
	if (!readType(parser, command, "set", setNames))
		return false;

	command.type.value = parseInt(command.type.value);

	// Set is just a test with a type.
	if (!parseTest(parser, command))
		return false;

	if (command.type.value == parseInt(findInList("pass", setNames)) &&
	    (command.aim.value == parseInt(findInList("x", testNames)) ||
	     command.aim.value == parseInt(findInList("%", testNames)) ||
	     command.aim.value == parseInt(findInList("00", testNames)) ||
	     command.aim.value == parseInt(findInList("ep", testNames)))) {
		openDialog("Error", "'set pass' doesn't currently work with the aims " +
		           "'x', '%', '00', and 'ep'.");
		return false;
	}

	return readAnyOptions(parser, command, "set", setOptions);
}

// Helper variable and functions to build subcommands in parseStipulation().

var commandBlackFirst = {command: "option",
                         text: "\noption",
                         type: {text: " FirstToMove", value: "FirstToMove"},
                         color: {text: " black", value: "black"}};

var commandNotationGap = {command: "option",
                          text: "\noption",
                          type: {text: " InitialMoveNumber",
                                 value: "InitialMoveNumber"},
                          number: {value: 1.5}};

// Length of series have already been subtracted by 0.5.
function commandPlay(type, length) {
	var plies = length * 2;

	var lengthStr = (type == "series" && length % 1 == 0.5) ?
	                (length + 0.5).toFixed(0) :
					length.toFixed(1);

	var id = parseInt(findInList(type, playNames));

	var ret = {command: "play",
	           text: "\nplay",
	           type: {text: " " + type,
	                  value: id},
	           length: {text: " " + lengthStr,
	                    value: plies}};
	return ret;
}

// Copy test from command.
function commandTestFrom(command) {
	var ret = {command: "test",
	           text: "\ntest",
	           aim: {text: " " + command.aim.text,
	                 value: command.aim.value}};

	if ("negated" in command)
		ret.negated = command.negated;

	if (command.aim.value == 0)
		ret.dia = command.dia;

	return ret;
}

// Copy test from command.
function commandSetTestReflexFrom(command) {
	var ret = commandTestFrom(command);
	ret.command = "set";
	ret.text = "\nset";
	ret.type = {text: " test",
	            value: parseInt(findInList("test", setNames))};
	ret.options = [{text: " reflex", value: findInList("reflex", setOptions)}];
	return ret;
}

function commandSet(type, negated, aim) {
	var ret = {command: "set",
	           text: "\nset",
	           type: {text: " " + type,
	                  value: parseInt(findInList(type, setNames))},
	           aim: {text: " " + aim,
	                 value: parseInt(findInList(aim, testNames))},
	           options: []};
	if (negated) {
		ret.negated = {text: " not",
		               value: true};
	}
	return ret;
}

function addClosureSubCommands(subcommands, closure) {
	for (var i = 0; i < closure.length; i++) {
		var c = closure.charAt(i).toLowerCase();
		if (c == "h")
			subcommands.push(commandPlay("help", 0.5));
		else if (c == "*")
			subcommands.push(commandPlay("pass", 0.5));
		else if (c == "s" || c == "r")
			subcommands.push(commandPlay("defense", 0.5));
		else
			return false;
	}

	return true;
}

function parseStipulation(parser, command) {
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		openDialog("Error", "Stipulation type is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();
	parser.curr++;

	// Identify the stipulation as prefix + closure + aimName.
	// Store the rest in lengthStr in case the length is part of the token.

	// prefix != null because the empty string is one of the options.
	var prefix = grabLongestMatch(token, prefixNames);
	var closure = grabClosure(token.substr(prefix.length));
	var aimName = grabLongestMatch(token.substr(prefix.length + closure.length),
	                               testNames);

	if (aimName == null) {
		openDialog("Error", "Stipulation '" + token + "' is unknown.");
		return false;
	}
	var aim = parseInt(findInList(aimName, testNames));
	var lengthStr = token.substr(prefix.length + closure.length +
	                             aimName.length);

	var separator2 = "";
	if (lengthStr == "") {
		// Grab next token.
		var token2 = parser.getCurrToken();
		if (token2 == "") {		// Undefined or empty.
			openDialog("Error", "Stipulation length is missing.");
			return false;
		}
		var separator2 = parser.getCurrSeparator();
		parser.curr++;

		lengthStr = token2;
	}

	// Store aim now, because it is copied in subcommand in different ways.
	command.prefix = {text: separator + prefix,
	                  value: prefix};	// Value used by addAntiParryIfNeeded.

	command.closure = {text: closure};	// Value unused.

	command.aim = {text: aimName,
	               value: aim};

	command.length = {text: separator2 + lengthStr};

	if (aim == 0) {
		if (!parseDia(parser, command, aimName))
			return false;
	}

	var isHelp = (prefix == "h" || prefix == "ph" || prefix == "ah" ||
	              prefix == "pah" || prefix == "u");
	var isSeries = (prefix == "ser-" || prefix == "phser-" ||
	                prefix == "ahser-" || prefix == "pahser-" ||
	                prefix == "puser-" || prefix == "auser-" ||
	                prefix == "pauser-");

	// Parse length.
	var re = /^\d*(|\.0|\.5)$/;
	if (!re.test(lengthStr)) {
		openDialog("Error", "Stipulation length '" + lengthStr + "' is wrong.");
		return false;
	}

	var length = parseFloat(lengthStr);
	if (isSeries) {
		re = /^\d*$/;
		var hasDecimals = !re.test(lengthStr);
		if (!hasDecimals)
			length -= 0.5;
	}

	var hasHalfMove = (length % 1 == 0.5);

	// Translate the compact conventional stipulations into elementary
	// commands.
	var subcommands = [];

	// Closure information.
	var lastHelpIndex = closure.lastIndexOf("h");
	if (lastHelpIndex == -1 && isSeries && !hasHalfMove)
		lastHelpIndex--;	// Series prefix is h*h*h*.
	var lastMove = closure.charAt(closure.length - 1);

	// Apply the rules given in an Article by Nicolas Dupont in cpb-13:
	// If a defensive move exists, then it is played by Black.
	// Otherwise, the last help move is played by White.
	var lastMoveBlack = (lastMove == "s" || lastMove == "r" ||
	                     (closure.length - lastHelpIndex) % 2 == 0);

	// Handle set test.
	if (prefix == "r" || ((isHelp || isSeries) && lastMove == "r"))
		subcommands.push(commandSetTestReflexFrom(command));
	else if (prefix == "u") {
		var subcommand = commandSet("test", false, "~");
		subcommand.options.push(
			{text: " second", value: findInList("second", setOptions)});
		subcommand.options.push(
			{text: " one", value: findInList("one", setOptions)});
		subcommands.push(subcommand);
	}
	else
		subcommands.push(commandSet("test", false, "~"));

	// Handle set pass only for stipulations that have pass moves.
	if (prefix == "phser-" || prefix == "pahser-" ||
	    prefix == "puser-" || prefix == "pauser-" ||
	    prefix == "ph" || prefix == "pah") {
		subcommands.push(commandSet("pass", true, "+"));
	}
	if (prefix == "ahser-" || prefix == "pahser-" ||
	    prefix == "auser-" || prefix == "pauser-" ||
	    prefix == "ah" || prefix == "pah") {
		subcommands.push(commandSet("pass", true, "!+"));
	}

	if (prefix == "puser-" || prefix == "auser-" || prefix == "pauser-") {
		subcommand = subcommands[subcommands.length - 1];
		subcommand.options.push(		
			{text: " unique", value: findInList("unique", setOptions)});
	}

	if (prefix == "ser-" ||
	    (subcommands.length == 0 && closure.indexOf("*") >= 0)) {
		subcommands.push(commandSet("pass", false, "~"));
	}

	// Add prefix subcommands.
	if (prefix == "") {
		if (closure.length != 0) {
			openDialog("Error",
			           "Closure must be empty in direct or PG genres.");
			return false;
		}

		if (aim == 0) {
			// Empty-dia is treated as a help game without
			// any blackFirst convention for integer length.
			if (length > 0)
				subcommands.push(commandPlay("help", length));
		}
		else {
			// For everything else, Empty-X is direct-X.
			if (length != 1) {
				openDialog("Error", "Direct stipulation length must be " +
				           "exactly 1 in Jacobi.");
				return false;
			}
			subcommands.push(commandPlay("help", 0.5));	
		}
		subcommands.push(commandTestFrom(command));
	}
	else if (prefix == "s" || prefix == "r") {
		if (closure.length != 0) {
			openDialog("Error", "Closure must be empty in 's' or 'r' genres.");
			return false;
		}
		if (length != 1) {
			openDialog("Error", prefix +
			           " length must be exactly 1 in Jacobi.");
			return false;
		}

		subcommands.push(commandPlay("help", 0.5));
		subcommands.push(commandPlay("defense", 0.5));
		subcommands.push(commandTestFrom(command));
	}
	else if (isHelp) {
		var prefixLength = length - closure.length * 0.5;

		if (hasHalfMove && prefix != "u")
			subcommands.push(commandNotationGap);
//		alert("h = " + hasHalfMove + ", l = " + lastMoveBlack + ", c " + closure.length + " , l = " + lastHelpIndex);
		if (hasHalfMove == lastMoveBlack && prefix != "u")
			subcommands.push(commandBlackFirst);

		if (prefixLength < 0) {
			openDialog("Error", token + " length must be longer.");
			return false;
		}

		if (prefixLength > 0)
			subcommands.push(commandPlay("help", prefixLength));
		if (!addClosureSubCommands(subcommands, closure)) {
			openDialog("Error", "Problem with closure '" + closure + "'.");
			return false;
		}
		subcommands.push(commandTestFrom(command));
	}

	else if (isSeries) {
		// This follows Popeye. Magazines sometimes do length-1 for ser-hs,
		// which is consistent with length-0.5 in hs, but inconsistent with
		// ser-h and ser-s where it should then be length-0.5.
		var prefixLength = length;

		if (prefixLength < 0.5) {
			openDialog("Error", token + " length must be longer.");
			return false;
		}

		if (lastMoveBlack == ((prefixLength*2 + closure.length) % 2 != 0) &&
		    aim != 0) {
			subcommands.push(commandBlackFirst);
		}

		subcommands.push(commandPlay("series", prefixLength));
		if (!addClosureSubCommands(subcommands, closure)) {
			openDialog("Error", "Problem with closure '" + closure + "'.");
			return false;		
		}
		subcommands.push(commandTestFrom(command));
	}
	else {
		openDialog("Error", "Unknown stipulation prefix '" + prefix + "'.");
		return false;
	}

	command.subcommands = subcommands;

	return true;
}

function addConditionToList(list, cond, prefix) {
	var parts = cond.split(": ");
	var parts2 = parts[0].split("_");

	if (prefix == "") {
		// Only accept conditions without prefix.
		if (parts2.length == 1)
			list.push(cond);
	}
	else {
		// Only accept conditions with matching prefix.
		if (parts2.length == 2 && prefix == parts2[0])
			list.push(parts2[1] + ": " + (parts.length == 1 ? cond : parts[1]));
	}
}

function conditionBase(name) {
	var i = name.indexOf("_");
	return (i >= 0) ? name.substr(0, i) : name;
}

function conditionTail(name) {
	var i = name.indexOf("_");
	return (i >= 0) ? name.substr(i + 1) : name;
}

// Condition is a space separated list. Returns an array of distinct prefixes.
function addConditionBases(set, cond) {
	var parts = cond.split(" ");
	for (var i = 0; i < parts.length; i++) {
		var prefix = conditionBase(parts[i]);
		if (set.indexOf(prefix) == -1)
			set.push(prefix);
	}
}

function parseCondition(parser, command) {
	command.conditions = [];

	var prefixSet = [];

	while (true) {
		var token = parser.getCurrToken();
		if (token == "")
			break;	// Ignore undefined or whitespace after last keyword.
		var separator = parser.getCurrSeparator();

		var cond = "unknown";

		// First priority: prefixed conditions.
		// Only allow a suffix after a prefix set of size 1, otherwise
		// if we allow "MustCheck Ultra", then it's too hard to break it
		// apart because the components of two tokens must be reorganized.
		if (prefixSet.length == 1) {
			var list = [];
			for (var i = 0; i < conditionNames.length; i++)
				addConditionToList(list, conditionNames[i], prefixSet[0]);
			for (var i = 0; i < comboNames.length; i++)
				addConditionToList(list, comboNames[i], prefixSet[0]);
			for (var i = 0; i < otherNames.length; i++)
				addConditionToList(list, otherNames[i], prefixSet[0]);

			cond = findInList(token, list);
			if (cond == "ambiguous") {
				openDialog("Error",
				           "Condition '" + token + "' is " + cond + ".");
				return false;
			}
		}

		// Second priority: non-prefixed conditions.
		if (cond == "unknown") {
			var list = [];
			for (var i = 0; i < conditionNames.length; i++)
				addConditionToList(list, conditionNames[i], "");
			for (var i = 0; i < comboNames.length; i++)
				addConditionToList(list, comboNames[i], "");
			for (var i = 0; i < otherNames.length; i++)
				addConditionToList(list, otherNames[i], "");

			cond = findInList(token, list);

			if (cond == "ambiguous") {
				openDialog("Error",
				           "Condition '" + token + "' is " + cond + ".");
				return false;
			}
		}

		if (cond == "unknown")
			break;	// Current token is not a cond.

		command.conditions.push({text: separator + token,
								 value: cond});
		parser.curr++;

		// Update prefix set.
		prefixSet = [];
		addConditionBases(prefixSet, cond);
	}

	return true;
}

// Parse constraints.
// We're severe about letter case because bP is a black pawn, and BP is a
// berolina pawn.
function parseConstraints(parser, command) {
	var constraints = [];
	while (true) {
		var token = parser.getCurrToken();

		// Grab an initial piece description.
		var re = /^([A-Z]{1,2})([a-h\u00E7\u00E9][1-8])(.*)$/;
		var results = re.exec(token);
		if (!results)
			break;	// Maybe we're done.

		var rest = results[3];

		var piece = pieceLetters[parser.language].indexOf(results[1]);
		if (piece < 0) {
			openDialog("Error", "Unknown piece '" + results[1] + "'.");
			return false;
		}

		var square = squareFromString(results[2]);
		if (square < 0) {
			// This should not happen because of regex.
			openDialog("Error", "Square '" + results[2] + "' is wrong.");
			return false;
		}

		var value = {
			// Mandatory fields only.
			initial_piece: piece,
			origin: square,
			phases: []
		};

		// Grab phases.
		while (true) {
			var re = /^([\-x~*>!])(.*)$/;
			var results = re.exec(rest);
			if (!results)
				break;

			var type = results[1];
			rest = results[2];

			re = new RegExp("^((?:\\[[a-h\\u00E7\\u00E9][1-8]])?)" +
			                "([a-h\\u00E7\\u00E9][1-8])(@?)" +
			                "((?:=[a-z]?[A-Z]{1,2})?)(.*)$");
			var results = re.exec(rest);
			if (!results) {
				if (type == "-" || type == "~") {
					value.phases.push({
						type: type.charCodeAt(0),
						open_ended: true,
						captured_origin: -1,
						square: -1,
						// TODO: get rid of this when won't be testing v0.6.7
						// any longer.
						or_below: false,
						piece: -1
					});
					break;
				}
				else {
					openDialog("Error", "Doesn't understand '" + type + rest +
					           "' .");
					return false;		
				}
			}

			var captured_origin = -1;
			if (results[1] != "") {
				captured_origin = squareFromString(results[1].substr(1, 2));
				if (captured_origin < 0) {
					// This should not happen because of regex.
					openDialog("Error", "Square '" + results[1] +
					           "' is wrong.");
					return false;
				}
				if (type != "x" && type != "*") {
					openDialog("Error", "Captured origin '" + results[1] +
					           "' is only allowed with move types " +
							   "'x' and '*' .");
					return false;
				}
			}

			var square = squareFromString(results[2]);
			if (square < 0) {
				// This should not happen because of regex.
				openDialog("Error", "Square '" + results[2] + "' is wrong.");
				return false;
			}

			var or_below = (results[3] == "@");
			rest = results[5];

			var piece = -1;
			if (results[4] != "") {
				re = /^=([a-z]?)([A-Z]{1,2})$/;
				var results2 = re.exec(results[4]);
				if (!results2) {
					// This should not happen.
					openDialog("Error", "Doesn't understand '" + results[4] +
					           "' .");
					return false;
				}

				var color = 0;
				if (results2[1] == "")
					color = 0;
				else if (results2[1] == pieceColorNames[parser.language][0][0])
					color = WHITE;
				else if (results2[1] == pieceColorNames[parser.language][1][0])
					color = BLACK;
				else {
					openDialog("Error", "Doesn't understand color '" +
					           results2[1] + "' .");
					return false;
				}

				piece = pieceLetters[parser.language].indexOf(results2[2]);
				if (piece < 0) {
					openDialog("Error", "Unknown piece '" + results2[2] + "'.");
					return false;
				}
				piece += color;
			}

			value.phases.push({
				type: type.charCodeAt(0),
				open_ended: false,
				captured_origin: captured_origin,
				square: square,
				or_below: or_below,
				piece: piece
			});
		}

		// Add optional field: min_moves, max_moves, min_rebirths, max_rebirths.
		if (rest != "") {
			// Maybe rest is a min_move or a move interval.
			// min_rebirths and max_rebirths also possible.
			// (?: [...] ) is a non-capturing group, so we can
			// question-mark it.
			var re = /^\(((?:\d+)?)((?:\.\.\d+)?)r?((?:\d+)?)((?:\.\.\d+)?)\)$/;
			var results = re.exec(rest);
			if (!results) {
				openDialog("Error", "Doesn't understand '" + rest + "' .");
				return false;
			}
			if (results[1] != "")
				value.min_moves = parseInt(results[1]);
			if (results[2] != "")
				value.max_moves = parseInt(results[2].substr(2));
			if (results[3] != "")
				value.min_rebirths = parseInt(results[3]);
			if (results[4] != "")
				value.max_rebirths = parseInt(results[4].substr(2));
		}

		var separator = parser.getCurrSeparator();
		constraints.push({text: separator + token, value: value});

		parser.curr++;
	}

	command.constraintsOr.push({text: "", value: constraints});

	return true;
}

function parseConstraintsOr(parser, command) {
	command.constraintsOr = [];

	while (true) {
		if (!parseConstraints(parser, command))
			return false;

		var token = parser.getCurrToken();
		if (token == "")
			return true;

		var type = findInList(token, constraintsOptions);
		if (type == "or") {
			var separator = parser.getCurrSeparator();
			var last = command.constraintsOr[command.constraintsOr.length - 1];
			last.text = separator + token;
			parser.curr++;
		}
		else
			return true;
	}
}

// Returns true on success.
function parseCommand(parser, commands) {
	var separator = parser.getCurrSeparator();
	var token = parser.getCurrToken();

	var command;
	if (token == "")	// Empty whitespace at the end.
		command = "ignore";
	else {
		// From token to command.
		command = findInList(token, commandNames);
		if (command == "unknown" || command == "ambiguous") {
			openDialog("Error", "Command '" + token + "' is " + command + ".");
			return false;
		}
	}

	var commandObj = {text: separator + token,
					  command: command};
	commands.push(commandObj);
	parser.curr++;

	if (command == "ignore")
		return true;

	if (command == "englishS" || command == "englishN" ||
	    command == "french" || command == "german") {
		parser.language = command;

		// Turn it into a "language" command so we don't have so many
		// different commands.
		commandObj.command = "language";
		commandObj.type = command;
		return true;
	}
	else if (command == "comment") {
		parseComment(parser, commandObj);
		commandObj.command = "ignore";	// Solver doesn't need to know.
		return true;
	}
	else if (command == "forsyth" || command == "pieces") {
		if (parser.seenDia) {
			// Someday I might allow a position to be specified with a series
			// of "forsyth" and "pieces" commands like Popeye, but right now
			// it's not working, so make sure we don't treat the second
			// command in a series as an initial position.
			openDialog("Error", "Misplaced command '" + token + "'.");
			return false;
		}
		if (parser.seenInitialPosition) {
			openDialog("Error", "Repeated initial position declaration ('" +
			           token + "').");
			return false;
		}
		parser.seenInitialPosition = true;

		if (command == "forsyth")
			return parseForsyth(parser, commandObj);
		else
			return parsePieces(parser, commandObj);
	}
	else if (command == "option")
		return parseOption(parser, commandObj);
	else if (command == "play")
		return parsePlay(parser, commandObj);
	else if (command == "test")
		return parseTest(parser, commandObj);
	else if (command == "set")
		return parseSet(parser, commandObj);
	else if (command == "stipulation")
		return parseStipulation(parser, commandObj);
	else if (command == "condition")
		return parseCondition(parser, commandObj);
	else if (command == "constraints")
		return parseConstraintsOr(parser, commandObj);
	else {
		openDialog("Error", "Command '" + token + "' is " + command +
		           " but has no parser.");
		return false;
	}
}

function textFromCommands(commands) {
	var ret = "";
	for (var i = 0; i < commands.length; i++) {
		var command = commands[i];
		ret += command.text;
		if (command.command == "forsyth" || command.command == "pieces")
			ret += command.position.text;
		else if (command.command == "option") {
			ret += command.type.text;
			if (command.type.value == "FirstToMove")
				ret += command.color.text;
			else if (command.type.value == "InitialMoveNumber")
				ret += command.number.text;
		}
		else if (command.command == "play") {
			ret += command.type.text + command.length.text;
		}
		else if (command.command == "test" ||
		         command.command == "set" ||
		         command.command == "stipulation") {
			if (command.command == "set")
				ret += command.type.text;

			if ("negated" in command)
				ret += command.negated.text;

			if (command.command == "stipulation")
				ret += command.prefix.text + command.closure.text;

			ret += command.aim.text;

			if (command.command == "stipulation")
				ret += command.length.text;

			if (command.aim.value == 0) {	// dia
				ret += command.dia.text + command.dia.position.text;
				for (var j = 0; j < command.dia.options.length; j++) {
					ret += command.dia.options[j].text;
					if (command.dia.options[j].value == "Add")
						ret += command.dia.addList.text;
				}
			}

			if (command.command == "set") {
				for (var j = 0; j < command.options.length; j++)
					ret += command.options[j].text;
			}
		}
		else if (command.command == "condition") {
			for (var j = 0; j < command.conditions.length; j++)
				ret += command.conditions[j].text;
		}
		else if (command.command == "constraints") {
			for (var j = 0; j < command.constraintsOr.length; j++) {
				var constraints = command.constraintsOr[j];
				for (var k = 0; k < constraints.value.length; k++)
					ret += constraints.value[k].text;
				ret += constraints.text;	// Possible "or" keyword.
			}
		}
	}
	return ret;
}

function buildConditionCheckboxes(list, selectorPrefix) {
	var $fieldsetLeft = $(selectorPrefix + "-left");
	var $fieldsetRight = $(selectorPrefix + "-right");
	var html = "";
	var lines = 0;

	for (var i = 0; i < list.length; i++) {
		var parts = list[i].split(": ");
		var name = parts[0];
		var index = name.indexOf("_");

		if (index == -1 && i > 0) {
			// Flush html block.
			var blockMidpoint = i - lines/2;

			if (blockMidpoint <= list.length / 2)
				$fieldsetLeft.append(html);
			else
				$fieldsetRight.append(html);
			html = "";
			lines = 0;
		}

		var shortName = false;
		var displayName;
		if (index == -1)
			displayName = name;
		else {
			var name2 = name.substr(index + 1);
			if (name.substr(0, index) == "PromOnly")
				shortName = true;
			displayName = "&#8212;&#160;" + name2;
			// Good looking, but not in Firefox for Android.
//			displayName = "&#10551;&#160;" + name.substr(index + 1);
			// Too distracting.
//			displayName = "&#8627;&#160;" + name.substr(index + 1);
		}
		if (shortName)
			html += '<div style="display:inline-block; vertical-align:top;">'
		if (index == -1)
			html += '<label>'
		else
			html += '<label class="level2">'	// Currently has no effect.
		html += '<input id="' + name + '" type="checkbox">' +
		        displayName + '</label>\n';
		if (shortName)
			html += '</div>'
		lines++;
	}

	// Flush last block.
	$fieldsetRight.append(html);
}

// $.escapeSelector was introduced only in Jquery 3.0.
// http://fellowtuts.com/jquery/escaping-special-characters-in-jquery-selector/
function jqSelector(str) { 
	return str.replace(/([;&,\.\+\*\~':"\!\^#$%@\[\]\(\)=>\|])/g, '\\$1'); 
}

function refreshUIConditions() {
	$("div#conditions input").prop("checked", false);

	for (var i = 0; i < g_commands.length; i++) {
		if (g_commands[i].command == "condition") {
			var conditions = g_commands[i].conditions;
			for (var j = 0; j < conditions.length; j++) {
				// Click the elementary conditions.
				var parts = conditions[j].value.split(" ");
				for (var k = 0; k < parts.length; k++) {
					$("div#conditions input#" + jqSelector(parts[k])
					  ).prop("checked", true);
				}
				// Click the combo conditions.
				for (var k = 0; k < comboNames.length; k++) {
					var parts = comboNames[k].split(": ");
					if (conditions[j].value == parts[1]) {
						$("div#conditions input#" + jqSelector(parts[0])
						  ).prop("checked", true);
						break;
					}
				}
			}
		}
	}
	$("div#conditions input").checkboxradio("refresh");
}

function refreshTextArea() {
// https://stackoverflow.com/questions/10699210/jquery-mobile-growing-text-area-with-ajax-data
	$("textarea#text-representation").val(textFromCommands(g_commands))
	                                 .keyup();
}

$(document).one("pagebeforecreate", "#solver", function() {
	buildConditionCheckboxes(conditionNames, "fieldset#conditions");
	buildConditionCheckboxes(comboNames, "fieldset#combos");
});

$(document).one("pagecreate", "#solver", function() {
	var $page = $(this);	// "this" is the page for pagecreate.

	// Is the AppCache implementation of Firefox and/or Chrome severely
	// broken even in a simple case? The code below detects non-uniform
	// updates.
	var $version = $page.find("div[data-role='header'] h1");
	if ($version.html() != "Jacobi " + g_version)
		$version.append(" (misloaded, JavaScript is " + g_version + ")");

	$page.find("select#first-to-move").on("change", function(event, ui) {
		var colorE = $(event.target).val();

		// Color name in chosen language.
		var color = "";
		if (colorE == "white")
			color = pieceColorNames[g_language][0];
		else if (colorE == "black")
			color = pieceColorNames[g_language][1];

		var sep = color.indexOf(":");
		if (sep >= 0)
			color = color.substr(0, sep);

		var seen = 0;
		var insert_index = -1;

		// Traverse the commands to either delete the option or select
		// a good place to insert it.
		for (var i = 0; i < g_commands.length; i++) {
			if (g_commands[i].command == "option") {
				if (g_commands[i].type.value == "FirstToMove") {
					seen++;
					if (seen >= 2 || colorE == "unset") {
						removeCommand(g_commands, i);
						i--;
						continue;
					}
					else if (seen == 1) {
						g_commands[i].color =
							{text: " " + color, value: colorE};
					}
				}
			}
			// Just before these commands would be a good spot.
			if (insert_index == -1 && (
			    (g_commands[i].command == "option" &&
			     g_commands[i].type.value == "InitialMoveNumber") ||
			    g_commands[i].command == "play" ||
			    g_commands[i].command == "test" ||
			    g_commands[i].command == "set" ||
				g_commands[i].command == "stipulation")) {
				insert_index = i;
			}
		}

		// At the end would be a good spot.
		if (insert_index == -1)
			insert_index = g_commands.length;

		if (colorE != "unset" && seen == 0) {
			g_commands.splice(insert_index, 0, {
				text: (insert_index == 0 ? "" : "\n") + "option",
				command: "option",
				type: {text: " FirstToMove", value: "FirstToMove"},
				color: {text: " " + color, value: colorE}
			});
			if (insert_index == 0 && g_commands.length >= 2)
				g_commands[1].text = "\n" + g_commands[1].text;
		}

		refreshTextArea();
	});

	$page.find("input#initial-move-number").on("change", function(event, ui) {
		var token = $(event.target).val();
		var re = /^-?\d*(|\.0|\.5)$/;
		if (!re.test(token)) {
			openDialog("Error", "Initial move number '" + token +
			           "' is wrong.");
			return;
		}
		var number = parseFloat(token);

		var seen = 0;
		var insert_index = -1;

		// Traverse the commands to either delete the option or select
		// a good place to insert it.
		for (var i = 0; i < g_commands.length; i++) {
			if (g_commands[i].command == "option") {
				if (g_commands[i].type.value == "InitialMoveNumber") {
					seen++;
					if (seen >= 2 || token == "") {
						removeCommand(g_commands, i);
						i--;
						continue;
					}
					else if (seen == 1) {
						g_commands[i].number =
							{text: " " + token, value: number};
					}
				}
			}
			// Just before these commands would be a good spot.
			if (insert_index == -1 && (
			    g_commands[i].command == "play" ||
			    g_commands[i].command == "test" ||
			    g_commands[i].command == "set" ||
				g_commands[i].command == "stipulation")) {
				insert_index = i;
			}
		}

		// At the end would be a good spot.
		if (insert_index == -1)
			insert_index = g_commands.length;

		if (number != 0 && seen == 0) {
			g_commands.splice(insert_index, 0, {
				text: (insert_index == 0 ? "" : "\n") + "option",
				command: "option",
				type: {text: " InitialMoveNumber", value: "InitialMoveNumber"},
				number: {text: " " + token, value: number}
			});
			if (insert_index == 0 && g_commands.length >= 2)
				g_commands[1].text = "\n" + g_commands[1].text;
		}

		refreshTextArea();
	});

	$page.find("div#conditions input").on("change", function(event, ui) {
		var name = event.target.id;
		var checked = event.target.checked;

		// Is this a combo condition?
		var combo = "";
		for (var k = 0; k < comboNames.length; k++) {
			var parts = comboNames[k].split(": ");
			if (name == parts[0]) {
				combo = parts[1];
				break;
			}
		}

		var inserted = false;
		var insert_index = -1;

		var affectedBases = [];
		// Traverse the commands to either delete the condition or select
		// a good place to insert it.
		for (var i = 0; i < g_commands.length; i++) {
			if (g_commands[i].command == "condition") {
				var conditions = g_commands[i].conditions;
				var insert_subindex = -1;
				for (var j = 0; j < conditions.length; j++) {
					var cond = conditions[j].value;

					if (cond == combo) {
						// The user clicked a combo button and it matches
						// this condition.
						if (!checked) {	// We want it removed.
							conditions.splice(j, 1);
							j--;
						}
						else {
							if (DEBUG) {
								console.log("'" + combo +
								            "' is already in commands");
							}
						}
						continue;
					}

					var prefixSet = [];
					addConditionBases(prefixSet, cond);
					if (prefixSet.indexOf(conditionBase(name)) < 0)
						continue;	// Not matching.

					addConditionBases(affectedBases, cond);

					var parts = cond.split(" ");
					if (parts.length > 1 &&
					    (!checked || prefixSet.length > 1)) {
						// We need to break the special condition apart.
						conditions.splice(j, 1);
						var j2 = j;
						for (var k = 0; k < parts.length; k++) {
							conditions.splice(j2, 0, {
								text: " " + conditionTail(parts[k]),
						        value: parts[k]});
							j2++;
						}
						// Simpler to reprocess conditions[j].
						j--;
						continue;
					}

					if (!checked) {
						if (name == cond || name == conditionBase(cond)) {
							// We know the condition has only one part.
							conditions.splice(j, 1);
							j--;
						}
					}
					else {
						// Since there was a base match, this spot would work.
						insert_subindex = j+1;	// This spot would work.
					}
				}

				if (insert_subindex >= 0) {
					conditions.splice(insert_subindex, 0, {
						text: " " + conditionTail(name),
						value: name});
					inserted = true;
					break;
				}

				if (conditions.length == 0) {
					// Get rid of empty command.
					removeCommand(g_commands, i);
					i--;
				}
				insert_index = i+1;
			}
			else if (g_commands[i].command != "ignore") {
				// Just after the last command that we don't ignore would be a
				// good spot.
				insert_index = i+1;
			}
		}

		// At the end would be a good spot.
		if (insert_index == -1)
			insert_index = g_commands.length;

		if (checked && !inserted) {
			var conditions;
			if (combo != "")
				conditions = [{text: " " + name, value: combo}];
			else if (name == conditionBase(name))
				conditions = [{text: " " + name, value: name}];
			else {
				conditions = [{text: " " + conditionBase(name),
				               value: conditionBase(name)},
				              {text: " " + conditionTail(name),
							   value: name}];
			}

			g_commands.splice(insert_index, 0, {
				text: (g_commands.length == 0 ? "" : "\n") + "condition",
				command: "condition",
				conditions: conditions
			});
		}

		// TODO: see if the conditions with an affected base name can be joined
		// in a combo from comboNames.

		// Refresh conditions from scratch. Otherwise if we do it by delta and
		// there's a bug, then the UI will be out-of-sync.
		refreshUIConditions();

		refreshTextArea();
	});

	$page.find("textarea#text-representation")
	     .on("focus", function(event, ui) {
		$("button#apply-changes").removeClass("ui-state-disabled");
	});

	$page.find("button#apply-changes").on("vclick", function(event, ui) {
		$("button#apply-changes").addClass("ui-state-disabled");

		// Parse textarea.
		var text = $("textarea#text-representation").val();

		// Split the content into "separators + tokens" parts.
		var parts = [];
		var i = 0;
		while (i < text.length) {
			var pair = {separator: "", token: ""};
			// Grab separator.
			while (i < text.length) {
				var c = text.charAt(i);
				if (!isWhiteSpace(c))
					break;
				pair.separator += c;
				i++
			}
			// Grab token.
			while (i < text.length) {
				var c = text.charAt(i);
				if (isWhiteSpace(c))
					break;
				pair.token += c;
				i++
			}
			parts.push(pair);
		}
//		console.log(parts);

		// Parse parts into commands.
		var commands = [];
		var parser = {
			parts: parts,
			curr: 0,
			language: "englishS",
			seenInitialPosition: false,
			seenDia: false,

			getCurrToken: function() {
				return (this.curr < this.parts.length) ?
				       this.parts[this.curr].token : "";
			},

			getCurrSeparator: function() {
				return this.parts[this.curr].separator;
			}
		};

		var success = true;
		while (parser.curr < parser.parts.length) {
			success = parseCommand(parser, commands);
			if (!success)
				break;
		}
		if (DEBUG) {
			console.log("success = " + success + ", commands = ", commands);
			console.log(textFromCommands(commands));
			console.log(JSON.stringify(commands));
		}
		if (!success)
			return;

		// Update options in the UI. First, assume defaults.
		$("select#first-to-move").val("unset");
		$("input#initial-move-number").val("");

		for (var i = 0; i < commands.length; i++) {
			if (commands[i].command == "option") {
				if (commands[i].type.value == "FirstToMove") {
					$("select#first-to-move").val(commands[i].color.value);
//						Not needed: JQuery does the job.
//						.attr('selected', true)
//						.siblings('option').removeAttr('selected');
				}
				else if (commands[i].type.value == "InitialMoveNumber") {
					$("input#initial-move-number").val(
						commands[i].number.value);
				}
			}
		}
		$("select#first-to-move").selectmenu("refresh", true);
		$("div#conditions input").checkboxradio("refresh");

		g_commands = commands;
		g_language = parser.language;
		g_createdDepth = null;

		refreshUIConditions();
	});

	// Load problem from hash if it's there. Do this after the button bindings
	// above.
	if ("load" in g_params) {
		$("textarea#text-representation").val(g_params.load).keyup();
		$("button#apply-changes").trigger("vclick");
	}
});

$(document).on("pagebeforeshow", "#solver", function(event, ui) {
	var $page = $(this);
	$page.find("a.jqm-navmenu-link" ).on("click", function() {
		$page.find(".jqm-navmenu-panel").panel("open");
	});

	if (g_load_request != null) {
		$("textarea#text-representation").val(g_load_request).keyup();
		$("button#apply-changes").trigger("vclick");
		g_load_request = null;
	}
});
