//"use strict";

// Adapted from https://googlechrome.github.io/samples/service-worker/basic/

// Modification notice:
// - I changed PRECACHE and PRECACHE_URLS to correspond to my application.
// - I removed the RUNTIME cache which I don't want. Everything needed should
//   be in the PRECACHE.
// - I added some comments.

// Original copyright and license:

/*
 Copyright 2016 Google Inc. All Rights Reserved.
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
     http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
*/

// Names of the cache used in this version of the service worker.
// Change to v2, etc. when you update any of the local resources, which will
// in turn trigger the install event again.
const PRECACHE = "v0.7.5";

// A list of local resources we always want to be cached. The paragraph below
// explains why we're able to cache pages outside of the scope.
const PRECACHE_URLS = [
	"solver.html",
	"icon-128x128.png",
	"icon-192x192.png",
	"theme-custom.css",
	"../v0.7.5/chessboard.css",
	"../v0.7.5/general.js",
	"../v0.7.5/problem-editor.js",
	"../v0.7.5/solver.js",
	"../v0.7.5/game-editor.js",
	"../v0.7.5/engine-asmjs.js",
	"../v0.7.5/engine-wasm.js",
	"../v0.7.5/engine-wasm.wasm",
	"../v0.7.5/doc.html",
	"../libs/jquery-1.12.4.min.js",
	"../libs/jqm-1.4.5/jquery.mobile-1.4.5.min.css",
	"../libs/jqm-1.4.5/jquery.mobile-1.4.5.min.js",
	"../libs/jqm-1.4.5/images/ajax-loader.gif",
	"../libs/setImmediate-1.0.5.js",
	"../libs/bowser-1.9.4.min.js"
];

// From https://stackoverflow.com/questions/35780397/understanding-service-worker-scope#comment59275020_35780776
//
// "the scope determines which pages are controlled by the service worker.
// Once a page is controlled by a service worker, all HTTP requests originating
// from the page, regardless of the request URL, will trigger the service
// worker's fetch event"

// Do not cache the web manifest file using your service worker.
// https://www.c-sharpcorner.com/article/making-web-sites-look-like-native-apps-without-the-app-store/
//	"manifest-w3c.json",

// The following files are only used by very old broswers. They should be
// included in the Offline zip file, but it's not a big deal
// to break Offline AppCache for them. We're talking about
// Android 2.1, 2.2, 2.3; iOS Safari 3.2, 4.1, 4.3; Safari 4, 5.

//	"../libs/jqm-1.4.5/images/icons-png/arrow-l-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/arrow-r-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/arrow-u-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/back-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/bars-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/carat-d-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/carat-l-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/delete-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/forward-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/minus-white.png",
//	"../libs/jqm-1.4.5/images/icons-png/plus-white.png",

// The install handler takes care of precaching the resources we always need.
self.addEventListener('install', event => {
  event.waitUntil(
    caches.open(PRECACHE)
      .then(cache => cache.addAll(PRECACHE_URLS))
      .then(self.skipWaiting())
  );
});

// The activate handler takes care of cleaning up old caches.
self.addEventListener('activate', event => {
  event.waitUntil(
    caches.keys().then(cacheNames => {
      return cacheNames.filter(cacheName => (cacheName != PRECACHE));
    }).then(cachesToDelete => {
      return Promise.all(cachesToDelete.map(cacheToDelete => {
        return caches.delete(cacheToDelete);
      }));
    }).then(() => self.clients.claim())
  );
});

// The fetch handler serves responses for same-origin resources from a cache.
// If no response is found, respond from the network.
self.addEventListener('fetch', event => {
  // Skip cross-origin requests, like those for Google Analytics.
  if (event.request.url.startsWith(self.location.origin)) {
    event.respondWith(
      caches.match(event.request).then(cachedResponse => {
        if (cachedResponse)
          return cachedResponse;

        // If no response is found, respond from the network.
        return fetch(event.request);
      })
    );
  }
});
